# Python Macro Language for Dragon NaturallySpeaking
#   (c) Copyright 1999 by Joel Gould
#   Portions (c) Copyright 1999 by Dragon Systems, Inc.
#
# Pythonwin, trying things with dication grammar, only seems to work when
# exclusively activated.
#
# Use testDialogForDicationGrammar.py (run from pythonwin), to start the window
# which accepts dictation from this grammar
#
# dictated text is inserted in front of text that is already there
#
# written by: Quintijn Hoogenboom (QH software, training & advies), june 2006
#
#
"""
try with title "Title" (from testDialogForDicationGrammar.py) or
               "Script" having a new (non saved python script)
"""
import natlink
from natlinkcore import natlinkutils
import win32ui
from natlinkcore import nsformat 

dictObj = None
debugMode = 1
##wantedTitle = "PythonWin - [Script1]" 
wantedTitle = "Title" # belonging to testDialogForDicationGrammar

class VoiceDictation:
    dictObj = None
    ctrl = None
    # Initialization.  Create a DictObj instance and activate it for the
    # dialog box window.  All callbacks from the DictObj instance will go
    # directly to the dialog box.

    def initialize(self, hndle=None):
        if self.dictObj != None:
            D("dictObj already initialized")
            return
        self.__class__.dictObj = natlink.DictObj()
        Cwnd = win32ui.GetForegroundWindow()
        for i in range(15000, 16000):
            try:
                ctrl = Cwnd.GetDlgItem(i)  # edit box
            except:
                pass
            else:
                D("got ctrl: %s: %s"% (i, ctrl))
                self.__class__.ctrl = ctrl
                break
        else:
            D("could not find a ctrl")
            self.__class__.ctrl = None
        self.dictObj.setBeginCallback(self.onTextBegin)
        self.dictObj.setChangeCallback(self.onTextChange)
        self.dictObj.activate(hndle)
##        self.dictObj.activate(dlg.GetSafeHwnd())


    def onTextBegin(self,moduleInfo):
        D('win dict, text begin')
        self.updateState()

    # We get this callback when something in the dictation object changes
    # like text is added or something is selected by voice.  We then update
    # the edit control to match the dictation object.

    def onTextChange(self,delStart,delEnd,newText,selStart,selEnd):
        D('win dict, text change:%s, %s, %s, %s, %s'% ( delStart, delEnd, newText, selStart, selEnd))
        self.dictObj.setLock(1)
        if self.ctrl:
            self.ctrl.SetSel(delStart,delEnd)
            self.ctrl.ReplaceSel(newText)
            self.ctrl.SetSel(selStart,selEnd)
        else:
            sendkeys(newText)
        self.dictObj.setLock(0)

    def updateState(self):
        if self.ctrl:
            D("ctrl")
            text = self.ctrl.GetWindowText()
            selStart,selEnd = self.ctrl.GetSel()
        else:
            D("fake text")
            text = "aa bb cc dd"
            selStart, selEnd = len(text), len(text)

        visStart,visEnd = self.getVisibleRegion()
        if (visStart, visEnd) == (None, None):
            visStart, visEnd = selStart, selEnd
        self.dictObj.setLock(1)
        self.dictObj.setText(text,0,0x7FFFFFFF)
        self.dictObj.setTextSel(selStart,selEnd)
        self.dictObj.setVisibleText(visStart,visEnd)
        self.dictObj.setLock(0)
        D("dictObj set with text: %s"% text)

    def getWindowText(self):
        if self.ctrl:
            return self.ctrl.GetWindowText()
        else:
            return "aa bb cc"

    def getSel(self):
        if self.ctrl:
            return self.ctrl.GetSel()
        else:
            return 4, 5

        
    def getVisibleRegion(self):
        """Utility subroutine which calculates the visible region of the edit

         control and returns the start and end of the current visible region.
        """
        if self.ctrl:
            top,bottom,left,right = self.ctrl.GetClientRect()
            firstLine = self.ctrl.GetFirstVisibleLine()
            visStart = self.ctrl.LineIndex(firstLine)

            lineCount = self.ctrl.GetLineCount()
            lastLine = lineCount
            for line in range(firstLine+1,lineCount):
                charInLine = self.ctrl.LineIndex(line)
                left,top = self.ctrl.GetCharPos(charInLine)
                if top >= bottom:
                    break
                lastLine = line

            visEnd = self.ctrl.LineIndex(lastLine+1)
            if visEnd == -1:
                visEnd = len(self.ctrl.GetWindowText())
            return visStart,visEnd
        else:
            return None, None
    
    # Call this function to cleanup.  We have to reset the callback
    # functions or the object will not be freed.
        
    def terminate(self):
        if self.dictObj:
            self.dictObj.deactivate()
            self.dictObj.setBeginCallback(None)
            self.dictObj.setChangeCallback(None)
            self.__class__.dictObj = None
            self.__class__.ctrl = None

class CommandGrammar(natlinkutils.GrammarBase):

    gramSpec = """
        <OK> exported = OK | okay;
        <scratch> exported = scratch [that];
        <clear> exported = clear | start again;
    """

    def initialize(self):
        print('initializing/loading CommandGrammar belonging to Python dictatebox')
        self.load(self.gramSpec)
        self.state = None
        self.prevHandle = -1
        self.txt = None # becomes the text dialog
        # keep same as pythonwin_dict grammar:
        self.exclusive = 0 # test with 1 or 0!
        self.dictObj = VoiceDictation()

        
    def gotBegin(self,moduleInfo):
        winHandle = moduleInfo[2]
        if self.prevHandle == winHandle:
            return
        # terminate dictObj:
        self.dictObj.terminate()
        self.prevHandle = winHandle
        if natlinkutils.matchWindow(moduleInfo, 'pythonwin', wantedTitle):
            if not self.isActive():
                self.dictObj.initialize(hndle=winHandle)
                print('activate (exclusive: %s) Pythonwin command grammar with dictate box, handle: %s'% \
                      (self.exclusive, winHandle))
                self.activateAll(winHandle, exclusive=self.exclusive)
        elif self.isActive():
            print('deactivate Pythonwin command grammar with dictate box')
            self.dictObj.terminate()
            self.deactivateAll()
            self.txt = None

    def onTextBegin(self, param):
        print("ontextbegin: %s"% param)

    def gotResultsInit(self,words,fullResults):
        if not self.txt:
            Cwnd = win32ui.GetForegroundWindow()
            self.txt=Cwnd.GetDlgItem(15008)  # edit box

    def gotResults_OK(self, words, fullResults):
        print('heard command:  %s '% words)
        sendkeys("{tab}{enter}")

    def gotResults_scratch(self, words, fullResults):
        print('heard command:  %s '% words)
        i,j = self.txt.GetSel()
        if i < j:
            sendkeys("{backspace}")

    def gotResults_clear(self, words, fullResults):
        print('heard command:  %s '% words)
        self.txt.Clear()

##class DictGrammar(natlinkutils.DictGramBase):
##    def __init__(self):
##        natlinkutils.DictGramBase.__init__(self)
##
##    def initialize(self):
##        print 'initializing/loading DictGrammar!!'
##        self.load()
##        self.state = None
##        self.isActive = 0
##        self.prevHandle = -1
##        self.txt = None # becomes the text dialog
##        self.exclusive = 1  # test with 1 or 0!
##        self.totalText = ''
##        self.dictObj = VoiceDictation()
##
##        
##    def gotBegin(self,moduleInfo):
##        winHandle = moduleInfo[2]
##        if self.prevHandle == winHandle:
##            if self.dictObj:
##                self.onTextBegin()
##                
##            return
##        self.dictObj.terminate()
##        self.prevHandle = winHandle
##        if natlinkutils.matchWindow(moduleInfo, 'pythonwin', wantedTitle):
##            if not self.isActive:
##                if not self.dictObj:
##                    self.dictObj.initialize(ctrl=None, hndle=winHandle)
##
##                print 'activate (exclusive: %s) Pythonwin dictation grammar, handle: %s'% \
##                      (self.exclusive, winHandle)
##                self.activate(winHandle, exclusive=self.exclusive)
##                self.isActive = 1
##        elif self.isActive:
##            print 'deactivate Pythonwin dictation grammar'
##            if self.dictObj:
##                self.dictObj.terminate()
##            self.deactivate()
##            self.txt = None
##            self.isActive = 0
##
##
##    def gotResults(self, words):
##        self.numberOfLinesBelow = 3  # for s&s or printing
##        print 'heard dictation:  %s '% words
##        self.dictObj.setBeginCallback()
####        dir txt: ['Clear', 'Copy', 'CreateWindow', 'Cut', 'FmtLines',
####                  'GetFirstVisibleLine', 'GetLine', 'GetLineCount',
####                  'GetSel', 'LimitText', 'LineFromChar', 'LineIndex',
####                  'LineScroll', 'Paste', 'ReplaceSel', 'SetReadOnly', 'SetSel']
##        
####        for funcName in ['GetFirstVisibleLine', 'GetLine', 'GetLineCount',
####                     'LineIndex', 'LineFromChar']:
####            func = getattr(txt, funcName)
####            result = func()
####            print '%s: %s'% (funcName, result)
##        self.dictObj.onTextBegin()
##        
##        actualLine = txt.GetLine()
##        actualLineNumber = txt.LineFromChar()
##        ## with no selection on, the current line is txt.GetLine(), others are txt.GetLine(index)
##        ## the current line number = txt.LineFromChar()
##        ## in this example the visible region is shown, until maximum 3 lines below the current line
##        print 'current line: %s'% txt.GetLine()
##        print 'line index: %s'% txt.LineIndex()
##        print 'line fromchar: %s'% txt.LineFromChar()
##        print 'GetSel: %s, %s'% txt.GetSel()
##        maxLineNum = min(txt.GetLineCount(), actualLineNumber+self.numberOfLinesBelow)
##        for line in range(txt.GetFirstVisibleLine(), maxLineNum):
##            text = txt.GetLine(line)
##            print 'line %s: %s'% (line, text)
##            
##        actual,self.state = nsformat.formatWords(words,self.state)
##        txt.ReplaceSel(actual)
##        self.totalText += actual
        
def D(t):
    """debug text if debugMode is on"""
    if debugMode:
        print(t)
        


# standard stuff Joel (adapted for possible empty gramSpec, QH, unimacro)
dictGrammar = None
##dictGrammar = DictGrammar()
##dictGrammar.initialize()
cmdGrammar = CommandGrammar()
cmdGrammar.initialize()
print('cmdGrammar initialized')
##print 'dictGrammar initialized'

def unload():
    global dictGrammar, thisGrammar
    if dictGrammar:
        dictGrammar.unload()
    if cmdGrammar:
        cmdGrammar.unload()



# _spokenforms.py 
#  written by: Quintijn Hoogenboom (QH softwaretraining & advies)
#  June 2011/March 2022
#
#pylint:disable=C0115, C0116, R0912, R0914, R0915, R0911, C0209
"""This module contains a class spokenforms that maintains spoken forms
for numbers, thus making it possible to use spoken forms for all numbers lists
in Unimacro grammars which need numbers 

tested with test_spokenforms.py (in tests directory of the unimacro repository)
(with pytest)
"""
import re
import os
import os.path
import sys
import shutil
import string
import operator
from functools import reduce

from natlinkcore import natlinkstatus

from dtactions.unimacro import inivars
from dtactions.unimacro import utilsqh

class NumbersError(Exception):
    pass

status = natlinkstatus.NatlinkStatus()

# for generateMixedListOfSpokenForms:
reNonAlphaNumeric = re.compile(r'\W+')
reNumeric = re.compile(r'(\d+)')
reNumericUnderscore = re.compile(r'^([\d+]_+)')        # so 1_text results in text 
reNumericBracketsEnd = re.compile(r'\s*[(][0-9]+[)]\s*$') 
reLetterUnderscore = re.compile(r'^([a-zA-z]{1,2}[_])')  # so a_text only results in text, az_text also .
reVowelQuote = re.compile(r"([aeiou])'([sn])")
# for A or A. in spoken forms
reMatchUpperCaseLetter = re.compile(r'\b[A-Z]\b(?![.])')
reMatchUpperCaseLetterDot= re.compile(r'\b[A-Z]\.')
def addDot( match ):
    """for use in reMatchUpperCaseLetter"""
    return match.group() + '.'
def looseDot( match ):
    """for use in reMatchUpperCaseLetterDot"""
    return match.group()[0]

def fixSingleLetters(spoken, DNSVersion):
    """replace A by A. in Dragon <= 10 and A. in A in Dragon >= 11
    
    >>> fixSingleLetters("A. BC D", 10)
    'A. BC D.'
    >>> fixSingleLetters("C. DE. F", 11)
    'C DE. F'
    
    """
    if DNSVersion <= 10:
        m = reMatchUpperCaseLetter.search(spoken)
        if m:
            spoken= reMatchUpperCaseLetter.sub(addDot, spoken)
    else:
        # Dragon 11
        m = reMatchUpperCaseLetterDot.search(spoken)
        if m:
            spoken= reMatchUpperCaseLetterDot.sub(looseDot, spoken)
    return spoken


thisBaseDirectory = os.path.split(sys.modules[__name__].__dict__['__file__'])[0]

# special list names to be defined here:
number1to99stripped = list(range(1, 20)) +  list(range(20, 91, 10))


#####
##### get spokenforms.ini from baseDirectory or SampleDirectory into userDirectory:
status = natlinkstatus.NatlinkStatus()
# baseDirectory = status.getUnimacroDirectoryFromIni()
# if not baseDirectory:
#     baseDirectory = ""
#     #raise ImportError( 'no baseDirectory found while loading spokenforms.py, stop loading this module')
# sampleBases = [thisBaseDirectory.lower()]
# if thisBaseDirectory.lower() != baseDirectory.lower():
#     print 'actions module has ambiguous baseDirectory: |%s|, this: |%s|: take the latter one!'% (baseDirectory, thisBaseDirectory)
#     sampleBases.insert(0, baseDirectory.lower())
    
unimacroDirectory = status.getUnimacroDirectory().lower()
sampleBases = [unimacroDirectory]                   
sampleDirectories = [os.path.join(base, 'sample_ini') for base in sampleBases]
sampleDirectories = [p for p in sampleDirectories if os.path.isdir(p)]
      
if not sampleDirectories:
    print('\nNo Unimacro sample directory not found: %s\nCHECK YOUR CONFIGURATION!!!!!!!!!!!!!!!!\n')
#else:
#    print 'sample_directories: %s'% sampleDirectories
    
userDirectory = status.getUnimacroUserDirectory()

if not os.path.isdir(userDirectory):
    try:
        os.mkdir(userDirectory)
    except OSError:
        print('cannot make inifiles directory: %s'% userDirectory)
####

currentlanguage = None
inifile = None
ini = None

def resetSpokenformsGlobals():
    """utility function for unittest, to ensure there is a fresh start
    """
    #pylint:disable=W0603
    global currentlanguage, ini, inifile
    currentlanguage = ini = inifile = None
    
def checkSpokenformsInifile(language):
    """copy if needed the correct inifile into the userDirectory
    
    print message if no valid file found and return False 
    """
    #pylint:disable=W0603
    global currentlanguage, ini, inifile
    if language and language == currentlanguage and ini:
        return inifile # all OK
    ini = None
    currentlanguage = language
    filename = '%s_spokenforms.ini'% language
    if language == 'test':
        testDirectory = os.path.join(unimacroDirectory, 'unimacro_test', 'test_inifiles')
        print('test spokenforms.ini from %s'% testDirectory)
        inifile = os.path.join(testDirectory, '%s_spokenforms.ini'% language)
        print('spokenforms, test inifile: %s'% inifile)
    else:
        inifile = os.path.join(userDirectory, '%s_spokenforms.ini'% language)

    if not os.path.isfile(inifile):
        # now try to copy from the samples:
        print('---try to find spokenforms.ini file in old version (UserDirectory) or sample_ini directory')
        for sample in sampleDirectories:
            sampleinifile = os.path.join(sample, filename)
            if os.path.isfile(sampleinifile):
                print('---copy spokenforms.ini from\nsamples directory: %s\nto %s\n----'% (sampleinifile, inifile))
                shutil.copyfile(sampleinifile, inifile)
                if os.path.isfile(inifile):
                    break
                print('cannot copy sample spokenforms inifile to: "%s"'% inifile)
                inifile = None
                return None
        else:
            print('no valid sample "%s" file found in one of %s sample directories:\n|%s|'% \
                      (filename, len(sampleDirectories), sampleDirectories))
            return None
    # now assume valid inifile:
    try:
        ini = inivars.IniVars(inifile)
    except inivars.IniError:
        print('Error in spokenforms inifile: "%s"'% inifile)
        m = str(sys.exc_info()[1])

        print('message: %s'% m)
        print('\n\n===please edit %s (open by hand)'% inifile)
        #win32api.ShellExecute(0, "open", inifile, None , "", 1)    
    else:
        return inifile
    return None

def openInifile(inifilepath):
    #pylint:disable=W0603
    global inifile, ini
    if not os.path.isfile(inifilepath):
        ini = inifile = None
        print('no inifile spokenforms.ini found. Please repair')
        return None
    try:
        ini = inivars.IniVars(inifilepath)
    except inivars.IniError:
        
        print('Error in numbers inifile: %s'% inifilepath)
        m = str(sys.exc_info()[1])
        print('message: %s'% m)
        # pendingMessage = 'Please repair spokenforms.ini file\n\n' + m
        ini = None
        print('please edit %s (open by hand)'% inifilepath)
        return None
        #win32api.ShellExecute(0, "open", inifile, None , "", 1)
    else:
        return ini

def editSpokenForms(comingFrom=None, name=None, language=None):
    """show the spokenforms.ini file in a editor
    """
    if not language:
        print('editSpokenForms: call with language is required!')
        return
    
    inifilepath = checkSpokenformsInifile(language)
    if not inifilepath:
        print('editSpokenForms: no valid spokenforms inifile for language "%s" available'% language)
        return        
    if comingFrom:
        name=name or ""
        comingFrom.openFileDefault(inifilepath, name=name)
    else:
        print('inifile: ', inifilepath)
        print('please edit (open by hand): %s'% inifilepath)
    #win32api.ShellExecute(0, "open", inifile, None , "", 1)
    print('note: you need to restart Dragon after editing the spoken forms inifile.')
    

showSpokenForms = editSpokenForms

class SpokenForms:
    """maintain a (class wide) dict of numbers -> spoken forms and vice versa
    a language is required at __init__ time.
    
    Also characters (a, b, etc) to spoken and vice versa
    
    It is assumed only one language can be active at the same time, so the dicts
    are kept global in the class
    """
    n2s = {}
    s2n = {}
    char2spoken = {} # characters of radio alphabet possibly extended
    spoken2char = {}
    abbrev2spoken = {} # lists of spoken forms
    spoken2abbrev = {}
    ext2spoken = {}    # file extensions (without the dot) (section "extensions")
    spoken2ext = {}    
    punct2spoken = {}  # punctuation, from section punctuationreverse (other way round!)
    spoken2punct = {}
    
  
    language = None
    
    def __init__(self, language, DNSVersion=None):
        # global in this module, does not extra work if language did not change:
        checkSpokenformsInifile(language)
        self.DNSVersion = DNSVersion or 15
        
        if self.language is None or self.language != language:
            self.__class__.language = None
            if self.filldicts(language):
                self.__class__.language = language
                
                
    def filldicts(self, language):
        """fill the base dictionaries for the class
        """
        self.n2s.clear()
        self.s2n.clear()
        if ini is None:
            print('no inifile for numbers spoken forms')
            return False
        # section in spokenforms.ini file:
        section = "numbers"
        if not section in ini.get():
            print('no section in spokenforms.ini for language: %s'% section)
            return False
        for k in ini.get(section):
            v = ini.get(section, k)
            try:
                n = int(k)
            except ValueError:
                print('invalid entry in spokenforms.ini for language: %s\n' \
                      '%s = %s  (key must be a integer)'% (language, k, v))
                continue
            for splitChar in ",;|":
                if v.find(splitChar) > 0:
                    V = [s.strip() for s in v.split(splitChar)]
                    break
            else:
                V = [v]
            for w in V:
                self.s2n[w] = n
            self.n2s[n] = V
        
        section = 'alphabet'
        for k in ini.get(section):
            v  = ini.get(section, k)
            if not v:
                continue
            for splitChar in ",;|":
                if v.find(splitChar) > 0:
                    V = [s.strip() for s in v.split(splitChar)]
                    break
            else:
                V = [v]
            for w in V:
                self.spoken2char[w] = k
            self.char2spoken[k] = V
            
        section = 'abbrevs'
        for kk in ini.get(section):
            ## all convert to lowercase:
            k = kk.lower()
            v  = ini.get(section, k)
            
            if not v:
                v = ' '.join(l.upper() for l in k)
            for splitChar in ",;|":
                if v.find(splitChar) > 0:
                    V = [s.strip() for s in v.split(splitChar)]
                    break
            else:
                V = [v]
            V = [fixSingleLetters(s, self.DNSVersion) for s in V]
            for w in V:
                self.spoken2abbrev[w] = k
            self.abbrev2spoken[k] = V

        section = 'extensions'
        for k in ini.get(section):
            v  = ini.get(section, k)
            if not v:
                v = ' '.join(l.upper() for l in k)
            for splitChar in ",;|":
                if v.find(splitChar) > 0:
                    V = [s.strip() for s in v.split(splitChar)]
                    break
            else:
                V = [v]
            V = [fixSingleLetters(s, self.DNSVersion) for s in V]
            for w in V:
                # extensions can point back to more extensions, eg excel to .xls and to .xlsx
                self.spoken2ext.setdefault(w, []).append(k)
            self.ext2spoken[k] = V

        section = 'punctuationreverse'
        for k in ini.get(section):
            try:
                v  = ini.get(section, k)
            except inivars.IniError:
                print('Error in ["%s"] section of spokenforms.ini files'% section)
                m = str(sys.exc_info()[1])
                print('message: %s'% m)
                v = None
                
            if not v:
                continue
                # extensions can point back to more extensions, eg excel to .xls and to .xlsx
            self.spoken2punct[k] = v
            self.punct2spoken.setdefault(v, []).append(k)
        
        return True
    
    def correctLettersForDragonVersion(self, spoken):
        """convert A to A. in NatSpeak <= 10 and B. int B for Dragon 11
        """
        return fixSingleLetters(spoken, self.DNSVersion)
    
    def getSpokenFormsList(self, n):
        """return a list of spoken forms, also for numbers larger than filldicts gave
        
        this one goes one way, only fills n2s if not available yet. Larger numbers may be included
        for future references.
        """
        if n in self.n2s:
            return self.n2s[n]
        spokenforms = self.generateSpokenFormsFromNumber(n)
        if spokenforms:
            self.n2s[n] = spokenforms
            return spokenforms
        return None
    
    def getMixedCharactersList(self, ListOfChars=None):
        """return a list of strings with either the number or the spoken forms
        """
        if ListOfChars is None:
            ListOfChars = string.ascii_lowercase
        L = []
        for s in ListOfChars:
            if s in self.char2spoken:
                L.extend(self.char2spoken[s])
            else:
                L.append(s)
        return L
    
    def getMixedList(self, ListOfNumbers):
        """return a list of strings with either the number or the spoken forms
        """ 
        L = []
        for s in ListOfNumbers:
            i = int(s) # in case we started with a string
            spokenforms = self.getSpokenFormsList(i)
            if spokenforms:
                L.extend(spokenforms)
                for name in spokenforms:
                    # needed for retrieving the numbers
                    # spokenforms back to the numbers:
                    if name not in self.s2n:
                        self.s2n[name] = i
            else:
                print('no spoken forms found for %s'% i)
                L.append(str(i))
        return L

    def generateSpokenFormsFromNumber(self, n):
        """return a list of possible spoken forms strings
        """
        for i in 100, 1000, 0, 1:
            if not i in self.n2s:
                print('spokenforms.ini should have entry or entries for %s'%i)
                return [str(n)]
        hundred = self.n2s[100][0]
        thousand = self.n2s[1000][0]
        oh = self.n2s[0][0]
        one = self.n2s[1][0]
    
        s = str(n)

        if n < 100:
            print('should not come here for %s, all numbers below 100 should be in ini file'%n)
            return None
        if n < 1000:
            if n == 100:
                print('should %s take from n2s'% n)
                return s
            if s.endswith('00'):
                digitstrings = self.n2s[int(s[0])]
                return [d+' '+hundred for d in digitstrings]
            if s[1] == '0':
                # one oh three 103 etc
                sp1 = self.n2s[int(s[0])]
                sp2 = self.n2s[int(s[2])]
                spList =['%s %s %s'% (s1, oh, s2) for s1 in sp1 for s2 in sp2]
                if s[0] == '1':
                    sp1 = self.n2s[100]
                else:
                    sp1 = ['%s %s'% (d, hundred) for d in self.n2s[int(s[0])]]
                spList.extend(['%s %s'% (s1, s2) for s1 in sp1 for s2 in sp2])
                return spList
            if s[0] == '1':
                # hundred ... or  one ...
                sp = self.n2s[int(s[1:])]
                spList = ['%s %s'% (hundred, s) for s in sp]
                spList.extend(['%s %s'% (one, s) for s in sp])
                return spList
            sp1 = self.n2s[int(s[0])]
            sp2 = self.n2s[int(s[1:])]
            spList = ['%s %s %s'% (s1, hundred, s2) for s1 in sp1 for s2 in sp2]
            spList.extend(['%s %s'% (s1, s2) for s1 in sp1 for s2 in sp2])
            return spList
        if n < 10000:
            if n == '1000':
                print('should %s take from n2s'% n)
                return s
            if s.endswith('000'):
                digitstrings = self.n2s[int(s[0])]
                return [d+' '+thousand for d in digitstrings]
            if s[1:3] == '00':
                # one oh three 103 etc
                sp1 = self.n2s[int(s[0])]
                sp2 = self.n2s[int(s[3])]
                spList =['%s %s %s %s'% (s1, oh, oh, s2) for s1 in sp1 for s2 in sp2]
                if s[0] == '1':
                    sp1 = self.n2s[1000]
                else:
                    sp1 = ['%s %s'% (d, thousand) for d in self.n2s[int(s[0])]]
                spList.extend(['%s %s'% (s1, s2) for s1 in sp1 for s2 in sp2])
                return spList
                    
            if s[1] == '0':
                # thousand ... or  ten ...
                if s[0] == '1':
                    sp1 = self.n2s[1000]
                else:
                    sp1 = ['%s %s'% (d, thousand) for d in self.n2s[int(s[0])]]
                sp2 = self.n2s[int(s[2:])]
                spList = ['%s %s'% (s1, s2) for s1 in sp1 for s2 in sp2]
                
                # twenty eleven:
                sp1 = self.n2s[int(s[0:2])]
                spList.extend(['%s %s'% (s1, s2) for s1 in sp1 for s2 in sp2])
                return spList
            sp1 = self.n2s[int(s[0:2])]
            if s[2] == '0':
                sp2 = ['%s %s'% (oh, d) for d in self.n2s[int(s[3])]]
            else:
                sp2 = self.n2s[int(s[2:4])]
            spList = ['%s %s'% (s1, s2) for s1 in sp1 for s2 in sp2]
            return spList
        return None
    
    def generateMixedListOfSpokenForms(self, s):
        """return spoken forms with at number places spoken forms replaced
            use for filenames with numbers in it
            
        if name startswith number_ skip it optionally
        if name endswith (number) skip it optionally
            
        if name startswith letter_ skip it optionally
        
        This is a too complicated function!!! (QH, 2023-10-02)
            
        """
        # make number _ at start and numbers in brackets at end of name optional
        # by recursing this function
        # eg 1_testing and website test (2)
        #
        s = reVowelQuote.sub(r'\1\1\2', s)  # no doubling from foto's to fotoos
        t = reNumericUnderscore.sub('', s)
        t = reNumericBracketsEnd.sub('', t)
        t = reLetterUnderscore.sub('', t)

        if t == s:
            prevResult = None
        else:
            prevResult = self.generateMixedListOfSpokenForms(t)
            #print 'intermediate: %s'% prevResult   
            
        Result = None # end result
        result = []   # intermediate result
        
        # dutch: e acute = ee etc.
        s = utilsqh.unifyaccentedchars(s)
        if not s.isalnum():
            sList = []
            for c in s:
                if c.isalnum():
                    sList.append(c)
                else:
                    sList.append(" ")
            s = ''.join(sList)
            s = s.replace('  ', ' ')
        # 
        # s = reNonAlphaNumeric.sub(' ', s)
        s = reNumeric.split(s)
        s = [_f for _f in s if _f]
        for i, phrase in enumerate(list(s)):
            if phrase.find(' ') > 0:
                phraseList = phrase.split()
                s[i:i+1] = phraseList
        for item in s:
            item = item.strip()
            if not item:
                continue
            try:
                n = int(item)
            except ValueError:
                ## to all lowercase???:
                if item.lower() in self.abbrev2spoken:
                    result.append(self.abbrev2spoken[item.lower()])
                else:
                    result.append(item)
            else:
                spok = self.getSpokenFormsList(n)
                if spok:
                    result.append(spok)
        for item in result:
            if isinstance(item, list):
                if Result:
                    Result = ['%s %s'% (i,j) for i in Result for j in item]
                else:
                    Result = item
            else:
                if Result:
                    Result = ['%s %s'% (r, item) for r in Result]
                else:
                    Result = [item]
        if prevResult:
            Result = prevResult + Result
        ## all lower case (?)
        return Result
    
    def getDictOfMixedSpokenForms(self, Values):
        """return a dict with spoken forms (multiple) as keys, and the Values as values
        (in grammar excel)
        """
        if not Values:
            return {}
        D = {}
        for v in Values:
            L = self.generateMixedListOfSpokenForms(v)
            if not L:
                continue
            for key in L:
                if key in D:
                    if D[key] == v:
                        continue
                    print("warning, double entry %s (%s and %s), take latter"% (key, D[key], v))
                D[key] = v
        return D
    
    def getCharFromSpoken(self, spoken, originalList=None):
        """try to retrieve the called character from the dict spoken2char in this class

        check result with originalList, 
        """
        first = self.spoken2char.get(spoken, None)
        if originalList is None:
            return first
        if first in originalList:
            return first
        return None
    
    def getPunctuationFromSpoken(self, spoken, originalList=None):
        """try to retrieve the called character from the dict spoken2char in this class

        check result with originalList, which may be any sequence
        """
        first = self.spoken2punct.get(spoken, None)
        if originalList:
            if first in originalList:
                return first
        else:
            return first
        return None

    def getNumberFromSpoken(self, spoken, originalList=None, asStr=None):
        """try to retrieve the called number from the dict s2n in this class

        check result with originalList, or try to make a number from spoken
        if asInt == True, return as str, otherwise return a int

        passing a list of words is handled in natlinkutilsbj.        
        """
        first = self.s2n.get(spoken, None)
        if originalList:
            orig = list(map(int, originalList))
            if first is None:
                try:
                    first = int(spoken)
                except ValueError:
                    pass
            if first is not None and first in orig:
                if asStr:
                    first = str(first)
                return first
        else:
            # take result anyway:
            if first is not None:
                if asStr:
                    first = str(first)
                return first
            # try is spoken represents a int:
            try:
                n = int(spoken)
                if asStr:
                    return str(n)
                return n
            except ValueError:
                pass
        return None
    
    def filldictsAboveHundred(self, num):
        """make spoken forms for numbers above 100
        eg 360 returns ['three sixty', 'three hundred sixty']
        """
        if num <= 100:
            raise ValueError('numbers, fillDictsAboveHundred call with number > 100, not: %s'% num)
        if num > 1000:
            raise ValueError('numbers, fillDictsAboveHundred call with number <= 1000, not: %s'% num)
        prefixSection = 'prefixes'
        numbersSection = 'numbers'
        if num == 1000:
            thousand = ini.getList(prefixSection, 'thousand') or ['thousand']
            self.n2s[1000] = thousand
            for t in thousand:
                self.s2n[t] = 1000
            return None

        snum = str(num)
        self.n2s[num] = [snum]
        self.s2n[snum] = num
        return None
        ## TODO QH
        h, rest = num / 100, num%100
        hundred = ini.getList(prefixSection, 'hundred', None) or ['hundred']
        if rest:
            srest = str(rest)
            restspoken = ini.getList(numbersSection, srest, None)
            if not restspoken:
                snum = str(num)
                return [snum]
        else:
            restspoken = ['']
        if num == 100:
            s100 = str(100)
            hspoken = ini.getList(numbersSection, s100, None) or hundred
        elif h == 1:
            s100 = str(100)
            hspoken = ini.getList(numbersSection, s100, None) or hundred
        else:
            hundredspoken = hundred[0]
            numspoken = ini.getList(numbersSection, str(h), None)
            if numspoken:
                # numbers + hundred or numbers:
                hspoken = [n + ' ' + hundredspoken for n in numspoken] + numspoken
            else:
                return [str(num)]
        spokenlist = [(a+ ' ' + b).strip() for a in hspoken for b in restspoken]
        self.n2s[num] = spokenlist
        for sp in spokenlist:
            self.s2n[sp] = num
        return None
            
    def sortedByNumbersValues(self, grammarsList, valueSpokenDict=None):
        """return sorted list if this is a numbers list (sorted by the number value)
        
        if valueSpokenDict == True: return the dict, like {1: ['one'], }
        otherwise return only the spoken forms like [ 'one', ...]
        
        If list contains items that do not go back to a number, return None
        TODOQH ik snap hier niets meer van.
        """
        dec = {}
        for g in grammarsList:
            value = self.getNumberFromSpoken(g)
            if value is None:
                return None
            dec.setdefault(value, []).append(g)
        if valueSpokenDict:
            return dec
        return reduce(operator.add, list(dec.values()))
        

    ## function to give numerical list:
    def getNumberList(self, listName):
        """returns the list of numbers for inserting in a grammar
    
        The list can be predefined here, and even be specific for different
        language versions.
    
        If the list is not predefined, it is extracted here.
    examples:
        number1to3 gives: [1, 2, 3]
        0-20 gives: [0, 1, ..., 20]
        
        but number10to99 gives [10, 11, ..., 99]
        
    Note:
        0-360 or number1to360 gives [0, 10, ..., 360]
        """
        #print 'try to fill numbers list: "%s"'% listName
        try:
            return globals()[listName]
        except KeyError:
            pass
        
        # not found in predefined names, extract automatically:
        if listName[:6] == 'number':
            numbersSpec = listName[6:]
        elif listName[0] == 'n':
            numbersSpec = listName[1:]
        if numbersSpec.find('to') > 0:
            L = numbersSpec.split('to')
        elif '-' in numbersSpec:
            L = numbersSpec.split('-')
        if len(L) != 2:
            print('getNumbersList, (1) seems not to be a valid definition of a numbers list: %s'% listName)
            return []
        try:  
            n1 = int(L[0])
            n2 = int(L[1])
        except ValueError:
            print('getNumbersList, (1) seems not to be a valid definition of a numbers list: %s'% listName)
            return []
        
        if n1 == 0 and n2 == 0:
            print('getNumbersList, (2) seems not to be a valid definition of a numbers list: %s'% listName)
            return []
            
        if n1%10 == 0 and n2%10==0 and (n2 > 100 or n1 >= 10):
            L = list(range(n1, n2+1, 10))
        else:
            L = list(range(n1, n2+1))
        #return [str(i) for i in L]
        return L

def _test():
    #pylint:disable=C0415
    import doctest
    doctest.testmod()

if __name__ == '__main__':
    _test()



# (unimacro - natlink macro wrapper/extensions)))
# (c) copyright 2003 Quintijn Hoogenboom (quintijn@users.sourceforge.net)
#                    Ben Staniford (ben_staniford@users.sourceforge.net)
#                    Bart Jan van Os (bjvo@users.sourceforge.net)
#
# This file is part of a SourceForge project called "unimacro" see
# http://unimacro.SourceForge.net).
#
# "unimacro" is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, see:
# http://www.gnu.org/licenses/gpl.txt
#
# "unimacro" is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; See the GNU General Public License details.
#
# "unimacro" makes use of another SourceForge project "natlink",
# which has the following copyright notice:
#
# Python Macro Language for Dragon NaturallySpeaking
#   (c) Copyright 1999 by Joel Gould
#   Portions (c) Copyright 1999 by Dragon Systems, Inc.
#
# _enabledictation.py

# make applications select-and-say, by inserting the dictobj class.
# no command grammars are inserted, this was done in the _global_dictation.py project.
# the applications that are "caught" are defined in windowsparameters.py
# currently only the windows messaging trick for checking and changing the foreground control is used.
# 
#  written by: Quintijn Hoogenboom (QH softwaretraining & advies)
#  September 2014
# 
"""make more windows select-and-say

The class needed here is the VoiceDictation class, which intercepts the
VDct of NatSpeak,

When speech starts, the onTextBegin of the dictobj instance (of VoiceDictation)
which is self.dictobj is called.

The callback function onTextChange is called when dictation has been recognised (of self.dictobj)

inside the VoiceDictation instance there is self.dictObj, which is the link to
the (intercepted) dictation object.

"""
import win32api
import win32gui
import win32con
import time
import os
import sys
import types
import re
#print 'sys.path: %s'% sys.path
import unimacro.messagefunctions as mess
import windowparameters

defaultWaitTime = 0.2  # change to slow down or speed up actions.

import natlink
from dtactions.unimacro.unimacroactions import doAction as action
from dtactions.unimacro.unimacroactions import doAction as action

from natlinkcore import natlinkutils
from dtactions.unimacro import unimacroutils
from dtactions.unimacro import unimacroutils
import unimacro.natlinkutilsbj as natbj

from time import sleep  

class VoiceDictation:
    dictObj = None  # is going to hold the internal dictObj (from natlink)
    #ctrl = None
    # Initialization.  Create a DictObj instance 
    # activate if outside the special text window

    def initialize(self):
        if self.dictObj != None:
            D("dictObj already initialized")
            return
        self.__class__.dictObj = natlink.DictObj()
        #self.__class__.app = None
        #self.__class__.ctrl = None  # to be filled in getEditControl
        self.dictObj.setBeginCallback(self.onTextBegin)
        self.dictObj.setChangeCallback(self.onTextChange)
        self.WindowsParameters = windowparameters.PROGS
        self.appNames = list(self.WindowsParameters.keys())
        print('_enabledictation tries applications: %s'% self.appNames)
        
        self.activeApp = None
        #self.getEditControl()
        # updated in updateState:
        self.lastSel = (0,0)
        self.lastSelText = ""
        self.prevHndle = None
        self.ctrl = None # hndle of the foreground control inside an application
        self.app = None  # hndle of the current foreground application (from moduleInfo)
        self.dctactive = None


    def onTextBegin(self,moduleInfo):
        """at start of each utterance, like gotBegin for command grammars
        """
        hndle = moduleInfo[2]
        if hndle != self.prevHndle:
            self.prevHndle = hndle
            if self.dctactive:
                dct = self.dictObj
                dct.deactivate()
                self.dctactive = 0
                progInfo = unimacroutils.getProgInfo(moduleInfo)

            _progpath, prog, title, _toporchild, _classname, _hndle = progInfo
            print('changing app to: "%s", %s'% (prog, hndle))
            self.app, self.ctrl = None, None
            if prog in self.WindowsParameters:
                self.getEditControl(prog, hndle)
                if self.ctrl:
                    print('got edit control for "%s": %s'% (prog, self.ctrl))
                    self.app = prog
                else:
                    print('no edit control found for "%s"'% prog)
        if self.ctrl is None:
            return
        if not self.dctactive:
            dct = self.dictObj
            print('activate dictObj to %s'% self.ctrl)
            dct.activate(self.ctrl)
            self.dctactive = 1
            
        self.updateState()
        #            
        #            
        #    
        #if self.activeApp is None:
        #    return
        #
        #
        #dct = self.dictobj
        #dct.getEditControl() # also notices when app has been killed.
        #
        #if self.prevHndle == hndle:
        #    if dct.app and not dct.hasFocus:
        #        dct.updateState()
        #    return   # no change
        #if modInfo[1] == "Spell":
        #    #print "Spell window"
        #    return
        #self.prevHndle = hndle
        #dct.saveFocus() # loosing the focus...
        ## focus changed, so hasFocus state can be changed:
        #if not dct.app:
        #    print 'no dictation application ready: %s'% dct.app
        #    dct.dictObj.deactivate()
        #    self.activateSet(self.outsidefocusrules)
        #    return
        #dct.updateState()
        #hndle = win32gui.GetForegroundWindow()
        #if dct.app == hndle:
        #    if dct.hasFocus != 1:
        #        self.activateSet(self.alwaysrules, exclusive=0)
        #        dct.hasFocus = 1
        #        print 'got focus'
        #        dct.dictObj.deactivate()
        #else:
        #    if dct.hasFocus != 0:
        #        # lost focus now:
        #        print "loosing focus"
        #        self.activateSet(self.outsidefocusrules)
        #        dct.hasFocus = 0
        #        dct.dictObj.activate(0)
        #
        #pass  # gets controlled from CmdsGrammar.gotBegin!
        ##if not self.app:
        ##    self.getEditControl()
        ##if not self.app:
        ##    return
        ###D('win dict, text begin')
        ##self.updateState()

    def getEditControl(self, prog, appHndle):
        """get the control (self.ctrl) of the text window in the foreground
        
        prog = name of application (to be defined in self.WindowsParameters)
        appHndle = hndle of the application
        """
        # get application, if not found, set ctrl to None
        currentSel = None
        if self.ctrl:
            currentSel = mess.getSelection(self.ctrl)
        if currentSel and currentSel != (0,0):
            return
        
        # get controls via messages functions, elaborate here if we need client-server trick:
        W = self.WindowsParameters[prog]
        # test case with fixed handle, set in windowparameters.py:
        if W["controlhandle"]:
            self.ctrl = W["controlhandle"]
            return




        wantedText, wantedClass, selectionFunction = W["edittext"], W["editcontrol"], W["selectionfunction"]
        print('wantedText: "%s", wantedClass: "%s", selectionFunction: %s'% (wantedText, wantedClass, selectionFunction))
        ctrls =  mess.findControls(appHndle, wantedText, wantedClass, selectionFunction) # pass the relevant windows parameters, as dict
        print('ctrls for "%s": %s'% (prog, ctrls))
        # some special triggering in a difficult case:
        if ctrls:
            editHndle = ctrls[0]
            self.ctrl = editHndle
        else:
            self.ctrl = None
            

    def wait(self, waitTime=None):
        waitTime = waitTime or defaultWaitTime
        time.sleep(waitTime)
        

    def onTextChange(self,delStart,delEnd,newText,selStart,selEnd):
        D('onTextChange: %s, %s, |%s|, %s, %s'% ( delStart, delEnd, repr(newText), selStart, selEnd) )
        
        if not self.ctrl:
            D("do not have ctrl control")
            return
        print('onTextChange for %s'% self.ctrl)
        dct = self.dictObj
        dct.setLock(1)
        getFocus = (selEnd - selStart > 1) # at start of call
        if newText:
            if self.lastSel == (delStart, delEnd):    
                dct.setText(self.lastSelText, delStart, selEnd)  # empty now
            else:
                print('window changed, update again')
                self.updateState()
            self.insertText(delStart, delEnd, newText)


        else:
            # loosing scratch that info:
            self.scratchinfo = []
            if delStart < delEnd:
                print('selection to delete: %s, %s'% (delStart, delEnd))
            if selStart < selEnd:
                self.setSelection(selStart,selEnd) # only set the window...
            else:
                self.insertText(selStart, selEnd, newText)
                
                #selStart, selEnd, newText = self.adjustAtEndOfText(selStart, selEnd, "")
                #if newText:
                #else:
                #    self.setSelection(selStart,selEnd) # only set the window...
                    
        dct.setLock(0)
       

    def insertText(self, delStart, delEnd, newText):
        # adjust at start:
        dct = self.dictObj
        newText = newText.replace('\n', '')                         
        delStart, newText = self.adjustAtStartOfText(delStart, newText)
        # make visible cursor of 1 space at the end
        selStart = delStart + len(newText)
        selStart, selEnd, newText = self.adjustAtEndOfText(selStart, delEnd, newText)
        # now go in internal buffer:
        dct.setText(newText, delStart, delEnd)
        dct.setTextSel(selStart, selEnd)

        # now the window buffer (through mess functions):                
        D("setting: |%s| to pos %s, %s"% (newText, delStart, delEnd))
        self.setSelection(delStart, delEnd)
        self.replaceSelection(newText)
        D("setting selection to pos %s, %s"% (selStart, selEnd))
        self.setSelection(selStart,selEnd)
        #if getFocus:
        #    unimacroutils.SetForegroundWindow(self.app)
        if not self.scratchThatCommand:
            self.scratchinfo.append( [(selStart, selEnd), newText,
                                  (delStart, delEnd), self.lastSelText])
        
    def correctForNewlines(self, t):
        """change \n into \r\n or \r\n into \n
        depending on self.linesep
        """
        if t.find('\n') == -1:
            return t
        
        if self.linesep == '\n':
            if t.find('\r') >= 0:
                t = t.replace('\r', '')
        elif self.linesep == '\r\n':
            t = t.replace('\n', '\r\n')
            t = t.replace('\r\r\n', '\r\n')
        print('correctedForNewlines: %s'% repr(t))
        return t

    def updateState(self):
        
        print('do updateState for %s'% self.ctrl)
        dct = self.dictObj
        if self.ctrl:
            text = self.getWindowText()
            #text = text.replace('\r', '')   # only \n for internal buffer
            selStart,selEnd = self.getSelection()
            #D("updateState, ctrl: selStart, selEnd: %s, %s"% (selStart, selEnd))
        visStart,visEnd = self.getVisibleRegion()
        print('visStart, visEnd: %s, %s'% (visStart, visEnd))
        if (visStart, visEnd) == (None, None):
            visStart, visEnd = 0, len(text)
        print('visStart, visEnd: %s, %s'% (visStart, visEnd))
        dct.setLock(1)
        dct.setText(text,0,0x7FFFFFFF)
        dct.setTextSel(selStart,selEnd)
        dct.setVisibleText(visStart, visEnd)
        self.lastSel = (selStart, selEnd)
        self.lastSelText = text[selStart:selEnd]
        D("dictObj set with text: %s (visible: %s, %s)"% (len(text), visStart, visEnd))
        D("part of text (10 left from selStart %s, selEnd: %s): |%s|"% (selStart, selEnd, repr(dct.getText(min(0, selStart-10), 0x7FFFFFFF))))
        dct.setLock(0)
            
       
    def getSelection(self):
        """get the selection of the edit control
        
        return a 2 tuple
        """
        return mess.getSelection(self.ctrl)

    def setSelection(self, start, end):
        """change the selection of the edit control
        """
        mess.setSelection(self.ctrl, start, end)

    def getWindowText(self):
        """get the buffer from the edit control
        
        should go smarter in
        """
        #print 'getting buffer'hallo
        tList = mess.getEditText(self.ctrl)
        return ''.join(tList)
        
    def replaceSelection(self, output):
        """overwrite selection with output
        """
        mess.replaceEditText(self.ctrl, output)
 
        
    def getVisibleRegion(self):
        """Utility subroutine which calculates the visible region of the edit

         control and returns the start and end of the current visible region.
        """
        return None, None
        if self.ctrl:
            buf = self.getWindowText()
            return 0, len(buf)
            
            top,bottom,left,right = self.ctrl.GetClientRect()
            firstLine = self.ctrl.GetFirstVisibleLine()
            visStart = self.ctrl.LineIndex(firstLine)

            lineCount = self.ctrl.GetLineCount()
            lastLine = lineCount
            for line in range(firstLine+1,lineCount):
                charInLine = self.ctrl.LineIndex(line)
                left,top = self.ctrl.GetCharPos(charInLine)
                if top >= bottom:
                    break
                lastLine = line

            visEnd = self.ctrl.LineIndex(lastLine+1)
            if visEnd == -1:
                visEnd = len(self.ctrl.GetWindowText())
            return visStart,visEnd
        else:
            return None, None
    
    # Call this function to cleanup.  We have to reset the callback
    # functions or the object will not be freed.
        
    def terminate(self):
        if self.dictObj:
            self.dictObj.deactivate()
            self.dictObj.setBeginCallback(None)
            self.dictObj.setChangeCallback(None)
            self.__class__.dictObj = None
            self.ctrl = None
            self.app = None

  
    def wait(self, waitTime=None):
        """some short time after each command
        
        waits after actions with messagefunctions are in the wrapper functions
        of the VoiceDictation instance
        """
        self.dictobj.wait(waitTime)
        
         
# debug print, only if testMode is o
testMode = 1
def D(t):
    if testMode:
        print(t)

dictationObject = VoiceDictation()
dictationObject.initialize()

#dictationObject = None
def unload():
    if dictationObject:
        dictationObject.terminate()
        dictationObject = None




# (unimacro - natlink macro wrapper/extensions)
# (c) copyright 2003 Quintijn Hoogenboom (quintijn@users.sourceforge.net)
#                    Ben Staniford (ben_staniford@users.sourceforge.net)
#                    Bart Jan van Os (bjvo@users.sourceforge.net)
#
# This file is part of a SourceForge project called "unimacro" see
# http://unimacro.SourceForge.net).
#
# "unimacro" is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, see:
# http://www.gnu.org/licenses/gpl.txt
#
# "unimacro" is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; See the GNU General Public License details.
#
# "unimacro" makes use of another SourceForge project "natlink",
# which has the following copyright notice:
#
# Python Macro Language for Dragon NaturallySpeaking
#   (c) Copyright 1999 by Joel Gould
#   Portions (c) Copyright 1999 by Dragon Systems, Inc.
#
# _calculator.py 
#  written by: Quintijn Hoogenboom (QH softwaretraining & advies)
#  April 2016
# 
"""smart and reliable number dictation, used for calculator

Inside the calc program the commands can be made exclusive.

But you can also use this grammar in other windows, because python itself can
do the calculations for you.
Also see the page "number grammar" on the Unimacro site and
the grammars _number simple and _number extended.
"""
import copy
from dtactions.unimacro.unimacroactions import doAction as action

from natlinkcore import natlinkutils as natut
from dtactions.unimacro import unimacroutils
from dtactions.unimacro import unimacroutils
import unimacro.natlinkutilsbj as natbj

ancestor = natbj.IniGrammar
class ThisGrammar(ancestor):

    language = unimacroutils.getLanguage()
    normalRules = ['calcnormal', 'cancel']
    calcRules = ['calccalc', 'cancel'] # exclusive
    continueRules = ['calccontinue', 'cancel'] # exclusive

    #see grammar _numbers.py (or _numbers_simple.py or _numbers_extended.py)
    #for instructions on the (quite complicated) number rules
    number_rules = natbj.numberGrammarTill999[language] # hundreds, including a long string of digits

    name = "calculator"

    gramSpec = """
<calcnormal> exported = calculate (<number>|<operator>)+ [equals];
<calccalc> exported = [calculate] (<number>|<operator>)+ [equals];
<calccontinue> exported = equals | (<number>|<operator>)+;
<number> = <integer> | <float>;
<operator> = {operator};
<cancel> exported = Cancel [Calculator] | Calculator Cancel;
"""+number_rules+"""
    """
  
    def initialize(self):
        if not self.language:
            print("no valid language in grammar "+__name__+" grammar not initialized")
            return
        self.load(self.gramSpec)
        # if switching on fillInstanceVariables also fill numbers lists like {n1-9} or {number1to99}
        self.switchOnOrOff()
        self.inCalculation = False
        self.activeSet = None
        self.minus = ''
        self.number = ''
        self.prevModInfo = None
        self.prog = None
        self.exclusive = 0
        self.lastResult = None
        self.startwithLastResult = False
        
    def gotBegin(self, moduleInfo):
        if self.checkForChanges:
            self.checkInifile()
        if self.prevModInfo != moduleInfo:
            progInfo = unimacroutils.getProgInfo(modInfo=moduleInfo)
            self.prevModInfo = moduleInfo
            self.cancelMode()
            self.prog = progInfo.prog
            if self.prog == 'calc':
                self.activeSet = None
            else:
                if self.exclusive:
                    self.activeSet = None  # reinitialize non exclusive

        if self.inCalculation:
            if self.prog != 'calc':
                if not self.activeSet == self.continueRules:
                    self.activeSet = copy.copy(self.continueRules)
                    self.exclusive = 1
                    self.activateSet(self.activeSet, exclusive=self.exclusive)
                    print("calculation, exclusiveRules activated")
        else:
            if self.prog == 'calc':
                if not self.activeSet == self.calcRules:
                    self.activeSet = copy.copy(self.calcRules)
                    self.exclusive = 1
                    self.activateSet(self.activeSet, exclusive=self.exclusive)
                    print("calc, exclusive rules activated")
            else:
                if not self.activeSet == self.normalRules:
                    if self.activeSet:
                        print("calculation, activate normalRules again")
                    self.activeSet = copy.copy(self.normalRules)
                    self.exclusive = 0
                    self.activateSet(self.activeSet, exclusive=self.exclusive)

    def gotResultsInit(self, words, fullResults):
        # Step 2:
        # you can initialise variables here, eg self.minus = None, self.number = ''. Not necessary, but
        # in some grammars more appropriate than in the next function.
        if self.inCalculation:
            pass
            # print 'continue calculation %s'% words
        else:
            ## reset, to be sure:
            self.calculation = []
            self.hadNumber = self.hadOperator = False

        if self.exclusive:
            print('calculation exclusive, wait for number')
            self.waitForNumber('number')
        # starting with no lastresult:
        self.startwithLastResult = False

    def fillInstanceVariables(self):
        """fills the necessary instance variables
        records the unary or possibly unary operators
        
        """
        self.unaryLeftOperators = self.ini.getList('general', 'unary left operators')
        self.unaryRightOperators = self.ini.getList('general', 'unary right operators')
        self.dualOperators = self.ini.getList('general', 'dual operators')  ## may be incomplete, but for double functions, like minus


    def reset(self):
        pass

    def gotResults_calcnormal(self, words, fullResults):
        # step 3: specify the number you are waiting for
        # check if the word Minus is present, and then wait for the number named "number"
        # note in this example self.number is non existent until collectNumber has been visited.
        # self.minus will only be set if the rule number is visited.
        # print 'calcnormal or calccalc'
        if self.hasCommon(words, 'calculate'):
            if self.prog == 'calc':
                print('calculate, clear previous calculation')
                keystroke("{esc}")
            self.waitForNumber('number')
            self.minus = False
            self.inCalculation = True

        elif self.hasCommon(words, 'equals'):
            self.collectNumberAddToCalculation()
            self.calculation.append("=")
            self.doCalculation()
            return

    gotResults_calccalc = gotResults_calcnormal
  
  
    def gotResults_calccontinue(self, words, fullResults):
        # step 3: specify the number you are waiting for
        # check if the word Minus is present, and then wait for the number named "number"
        # note in this example self.number is non existent until collectNumber has been visited.
        # self.minus will only be set if the rule number is visited.
        if not self.inCalculation:
            raise Exception("_calculator, shoud be in a calculation right now")
        
        if self.hasCommon(words, 'equals'):
            self.collectNumberAddToCalculation()
            self.calculation.append("=")
            self.doCalculation()
            self.inCalculation = False
       
    def gotResults_number(self, words, fullResults):
        # this rule only catches with the word "minus"
        print('calculator: rule number, %s'% words)
        # self.waitForNumber('number')
        self.minus = not self.minus

    def gotResults_operator(self, words, fullResults):
        # step 3: specify the number you are waiting for
        # check if the word Minus is present, and then wait for the number named "number"
        # note in this example self.number is non existent until collectNumber has been visited.
        # self.minus will only be set if the rule number is visited.
        
        for w in words:
            self.collectNumberAddToCalculation()
            print('rule operator, hadOperator: %s, hadNumber %s'% (self.hadOperator, self.hadNumber))
            operator = self.getFromInifile(w, 'operator')
            operatorStripped = operator.strip()
    
            unaryLeft = w in self.unaryLeftOperators
            unaryRight = w in self.unaryRightOperators
            possiblyDual = w in self.dualOperators
            
            if self.hadNumber:
                if unaryRight:
                    pass
                elif possiblyDual or not unaryLeft:
                    self.hadNumber = False  # just set now
                else:
                    print('no unaryLeft operator expected: %s'% w)
                    continue
            else:
                if unaryLeft and not possiblyDual:
                    pass
                elif w == words[0]:
                    self.startwithLastResult = True

            if self.hadOperator:
                operator = operator.strip()

            self.calculation.append(operator)
            if not unaryRight:
                self.hadOperator = True

    def gotResults_cancel(self, words, fullResults):
        self.cancelMode()


    def gotResults(self,words,fullResults):
        # step 4, in gotResults collect the number (as a string):
        self.collectNumberAddToCalculation() # setting self.number, see self.waitForNumber above
        if self.calculation:
            if self.prog == 'calc':
                print('calculation sofar (calc): %s'% ''.join(self.calculation))
                self.doCalculation()
            else:
                print('continue calculation or finish with "equals" (or "cancel")')
                print('calculation sofar: %s'% ''.join(self.calculation))
                
 


    def collectNumberAddToCalculation(self):
        """collect the number, and add to self.calculation
        
        also wait again for new number...
        """
        self.collectNumber()
        self.hadNumber = False
        if self.number:
            # if not self.hadOperator:
            #     print 'no operator found, for new number: %s'% self.number
            self.hadNumber = True
            self.hadOperator = False
            if self.minus:
                self.number = '-' + self.number
            self.calculation.append(self.number)
            self.number = ''
            self.waitForNumber('number')
                      
    def doCalculation(self):
        if not self.calculation:
            print('doCalculation, no calculation ready')
            return
        if not self.inCalculation and self.prog != 'calc':
            print('not in a calculation')
            return

        if self.startwithLastResult:
            self.calculation.insert(0, str(self.lastResult))
        
        calculationpython = ''.join([c for c in self.calculation if c != "="])
        # place for python specific tricks:
        if calculationpython.find(',') >= 0:
            calculationpython = calculationpython.replace(',', '.')
        result = eval(calculationpython)
        self.lastResult = result

        if self.prog in ['calc']:
            if self.startwithLastResult:
                del self.calculation[0]  # take memory from calculator, not from this list...
                print('in calc, continuation of calculation, expect lastResult: %s'% self.lastResult)
            for c in self.calculation:
                keystroke(c.strip())

        elif self.prog in ['natspeak']:  # DragonPad
            print('%s = %s'% (calculationpython, result))
        elif self.prog in ['excel']:
            calculationExcel = [c.strip() for c in self.calculation if c != "="]
            calculationExcel = '=' + ''.join(calculationExcel) + "{tab}"
            keystroke(calculationExcel)
        else:
            print("calc: %s = %s"% (calculationpython, result))
            keystroke('%s'% result )
            
        self.inCalculation = False
        self.calculation = []

    def cancelMode(self):
        #print "end of oops exclusive mode", also called when microphone is toggled.
##        print 'resetting oops grammar'
        if self.inCalculation:
            print('%s, cancel exclusive mode'% self.name)
            self.activeSet = None
            self.deactivateAll()
            self.inCalculation = False
            self.lastResult = None
        self.calculation = []


# standard stuff Joel (adapted for possible empty gramSpec, QH, unimacro)
thisGrammar = ThisGrammar()
if thisGrammar.gramSpec:
    thisGrammar.initialize()
else:
    thisGrammar = None

def unload():
    #pylint:disable=W0603
    global thisGrammar
    if thisGrammar: thisGrammar.unload()
    thisGrammar = None
    
def changeCallback(type,args):
    # not active without special version of natlinkmain:
    if ((type == 'mic') and (args=='on')):
        return   # check WAS in natlinkmain...
    if thisGrammar:
        thisGrammar.cancelMode()
    
    



#
# Python Macro Language for Dragon NaturallySpeaking
#   (c) Copyright 1999 by Joel Gould
#   Portions (c) Copyright 1999 by Dragon Systems, Inc.
#
# natspeak.py
#
# grammar activated if a recent folders dialog is brought to the foreground,
# interaction with grammar _folders.
#
#pylint:disable=
from dtactions.unimacro import unimacroutils
from dtactions.sendkeys import sendkeys as keystroke
import unimacro.natlinkutilsbj as natbj



ancestor = natbj.IniGrammar
class ThisGrammar(ancestor):
    """grammar recent folder dialog
    
    only to be switched on if dialog is there, if dialog is gone, the grammar rules set is deactivated
    """    
    language = unimacroutils.getLanguage()
    name = "natspeak dialog"
    
    setRecentFolderDialog = ["chooserfd", "okcancelrfd"]
    
    gramSpec = """
<chooserfd> exported = choose {number};
<okcancelrfd> exported = OK | Cancel;
    """
    def initialize(self):
        if not self.language:
            print("no valid language in grammar "+__name__+" grammar not initialized")
            return
        self.load(self.gramSpec)
        print(f'loaded: {self.gramSpec}') 
        self.prevHndle = None

    def gotBegin(self,moduleInfo):
        """switch rfd (recent folders dialog) on
        
        if title matches what is provided in _folders grammar.
        """
        hndle = moduleInfo[2]
        if hndle == self.prevHndle:
            return
        self.prevHndle = hndle
        if unimacroutils.matchModule('natspeak'):
            # try to reach data from _folders grammar: 
            self.foldersGram = self.GetGrammarObject('folders')
            #print 'self.foldersGram: %s'% self.foldersGram
            if self.foldersGram is None:
                return
            dTitle = self.foldersGram.dialogWindowTitle
            dRange = self.foldersGram.dialogNumberRange
            if dTitle and unimacroutils.matchModule('natspeak', wantedTitle=dTitle, titleExact=1, caseExact=1):
                self.activateSet(self.setRecentFolderDialog)
                self.setNumbersList('number', dRange)
                return
        # other situations:
        self.deactivateAll()
        
    def gotResults_chooserfd(self, words, fullResults):
        """result from the recent folders dialog
        
        extract the number, then close the dialog and
        then call the relevant function from the folders grammar
        """
        wNumList = self.getNumbersFromSpoken(words) # returns a string or None
        if len(wNumList) != 1:
            print(f'natspeak dialog: error in command, no number found: {wNumList}')
            return
        chooseNum = wNumList[0]
        self.exitDialog()
        self.foldersGram.gotoRecentFolder(chooseNum-1)  # remember choose is 1 based, the list is 0 based 

    def gotResults_okcancelrfd(self, words, fullResults):
        print(f'dialog ok cancel: {words}, just close the dialog')
        self.exitDialog()

    def exitDialog(self):
        """finish the open dialog, normally by pressing enter"""
        dTitle = self.foldersGram.dialogWindowTitle
        if dTitle and unimacroutils.matchModule('natspeak', wantedTitle=dTitle, titleExact=1, caseExact=1):
            unimacroutils.rememberWindow()
            keystroke("{enter}") # exit dialog.
        unimacroutils.waitForNewWindow()
        
        
# standard stuff:
thisGrammar = ThisGrammar()
thisGrammar.initialize()

def unload():
    #pylint:disable=W0603
    global thisGrammar
    if thisGrammar:
        thisGrammar.unload()
    thisGrammar = None



#
# Python Macro Language for Dragon NaturallySpeaking
#   (c) Copyright 1999 by Joel Gould
#   Portions (c) Copyright 1999 by Dragon Systems, Inc.
#
# unittestNatlinkutilsqh.py
#   This script performs tests on the utilities in natlinkutilsqh.py, now unimacroutils.py
#   the site mechanism (of Qh private) in which modules of a website generating program
#   are tested specifically
#
#pylint:disable=C0209, E1101
import sys
import unittest
import os
import os.path
import natlink
from dtactions.unimacro import unimacroutils
import TestCaseWithHelpers

def getBaseFolder(globalsDict=None):
    """get the folder of the calling module.

    either sys.argv[0] (when run direct) or
    __file__, which can be empty. In that case take the working directory
    """
    globalsDictHere = globalsDict or globals()
    baseFolder = ""
    if globalsDictHere['__name__']  == "__main__":
        baseFolder = os.path.split(sys.argv[0])[0]
        print('baseFolder from argv: %s'% baseFolder)
    elif globalsDictHere['__file__']:
        baseFolder = os.path.split(globalsDictHere['__file__'])[0]
        print('baseFolder from __file__: %s'% baseFolder)
    if not baseFolder or baseFolder == '.':
        baseFolder = os.getcwd()
        print('baseFolder was empty, take wd: %s'% baseFolder)
    return baseFolder

thisDir = getBaseFolder(globals())
logFileName = os.path.join(thisDir, "testresult.txt")

#---------------------------------------------------------------------------
# These tests should be run after we call natConnect
# no reopen user at each test anymore..
# no default open window (open window will be the calling program)
# default .ini files pop up when you first run this test. just ignore them.
class UnittestNatlinkutilsqh(TestCaseWithHelpers.TestCaseWithHelpers):
    def setUp(self):
        natlink.natConnect()
        
    def tearDown(self):
        natlink.natDisconnect()
    
    def testMatchTitle(self):
        """test the matchTitle function
        
        """
        modInfo = natlink.getCurrentModule()
        title = modInfo[1]
        ## construct a testword and a otherCaseTestWord:
        testword = title[2:]
        otherCaseTestWord = testword.lower()
        if otherCaseTestWord == testword:
            otherCaseTestWord = testword.upper()
            if otherCaseTestWord == testword:
                raise ValueError(" TEST ERROR, test of matchTitle fails because of invalid otherCaseTestWord: %s"% otherCaseTestWord)

        ## default case:
        got = unimacroutils.matchTitle(testword, modInfo=modInfo)
        exp = unimacroutils.getProgName(modInfo)
        self.assert_equal(exp, got, "matchTitle should be OK with testword: %s and window title: %s"% (testword, title))
        got = unimacroutils.matchTitle(testword, modInfo=modInfo)
        exp = unimacroutils.getProgName(modInfo)
        self.assert_equal(exp, got, "matchTitle should be OK with otherCaseTestWord: %s and window title: %s"% (otherCaseTestWord, title))

        ## rare different cases, all less important:
        got = unimacroutils.matchTitle(otherCaseTestWord, modInfo=modInfo, caseExact=1)
        exp = False
        self.assert_equal(exp, got, "matchTitle should fail with caseExact True: testword: %s and window title: %s"% (otherCaseTestWord, title))

        ## should match exact title, False!
        got = unimacroutils.matchTitle(otherCaseTestWord, modInfo=modInfo, titleExact=1)
        exp = False
        self.assert_equal(exp, got, "matchTitle should fail with titleExact True: testword: %s and window title: %s"% (otherCaseTestWord, title))
        got = unimacroutils.matchTitle(testword, modInfo=modInfo, titleExact=1)
        exp = False
        self.assert_equal(exp, got, "matchTitle should fail with titleExact True: testword: %s and window title: %s"% (testword, title))

        

        ## now for the whole title:
        wholeTitle = title
        exp = unimacroutils.getProgName(modInfo)
        got = unimacroutils.matchTitle(wholeTitle, modInfo=modInfo)
        self.assert_equal(exp, got, "matchTitle should be OK with whole title %s\nand window title: %s"% (wholeTitle, title))
        otherWholeTitle = wholeTitle.upper()
        if wholeTitle == otherWholeTitle:
            otherWholeTitle = wholeTitle.lower()
            if wholeTitle == otherWholeTitle:
                raise ValueError(" TEST ERROR, test of matchTitle wholeTitle fails because of invalid otherWholeTitle: %s"% otherWholeTitle)

        got = unimacroutils.matchTitle(wholeTitle, modInfo=modInfo, titleExact=1)
        self.assert_equal(exp, got, "matchTitle should be OK with whole title %s\nand window title: %s"% (wholeTitle, title))
        got = unimacroutils.matchTitle(wholeTitle, modInfo=modInfo, titleExact=1, caseExact=1)
        self.assert_equal(exp, got, "matchTitle should be OK with whole title %s\nand window title: %s"% (wholeTitle, title))

        got = unimacroutils.matchTitle(otherWholeTitle, modInfo=modInfo, titleExact=1)
        self.assert_equal(exp, got, "matchTitle should be OK with whole title %s\nand window title: %s"% (wholeTitle, title))
  
        got = unimacroutils.matchTitle(otherWholeTitle, modInfo=modInfo, titleExact=1, caseExact=1)
        exp = False
        self.assert_equal(exp, got, "matchTitle should fail with whole title %s\nand window title: %s"% (wholeTitle, title))
        got = unimacroutils.matchTitle(otherWholeTitle, modInfo=modInfo, titleExact=1, caseExact=1)
        self.assert_equal(exp, got, "matchTitle should be OK with whole title %s\nand window title: %s"% (wholeTitle, title))
                
        ## now for a list of possible test words:
        ListOfTestWords = ["xxxyyy", testword]
        exp = unimacroutils.getProgName(modInfo)
        got = unimacroutils.matchTitle(ListOfTestWords, modInfo=modInfo)


        ListOfTestWords = ["xxxyyy", otherCaseTestWord]
        exp = False
        got = unimacroutils.matchTitle(ListOfTestWords, modInfo=modInfo, caseExact=1)
            
def log(t):
    print(t)

def run():
    log('starting unittestNatlinkutilsqh')
    # trick: if you only want one or two tests to perform, change
    # the test names to her example def test....
    # and change the word 'test' into 'tttest'...
    # do not forget to change back and do all the tests when you are done.
    suite = unittest.makeSuite(UnittestNatlinkutilsqh, 'test')
##    natconnectOption = 0 # no threading has most chances to pass...
    result = unittest.TextTestRunner().run(suite)
    print(result)
if __name__ == "__main__":
    run()



import D_

#Predefined for training purposes

Numbers = ["zero", "numeral", "0",'1','2','3','4','5','6','7','8','9','10','11','12','13',
      '14','15','16','17','18','19','20','25','30','35','40','45','50']
AppStarts = ["hyper-active",
    "Switch to", "Start TextPad", "Start BibPad", "Start Email", "Start TeX", "Start Studio New",
             "Start Netscape Communicator", "Start MATLAB", "Start APL", "Start M.S. Word",
             "Start Python Windows", "Start Bib Database", "Start FTP Pro",
             "Switch to Eudora"]
TeXFundamentals = ["Bib TeX mode", "stop Bib TeX mode", "TeX mode", "stop TeX mode",
                   "replace TeX signs",
                   'Bib Field','Next Field','Next Number','Next Pages','Bib Number',
                   'Get','get TeX command list','TeX','get TeX environment list','TeX around',
                   'Symbol','Term','Get Symbol list','Get Term list','Get Math Symbol list',
                   'Insert Equation','Inline Equation','Insert','Search','Get reference list',
                   'Search reference list','Search All In Title','Search And','Get All In Title',
                   'Get And','Search Some In Title','Search Or','Get Some In Title','Get Or'
                    ]                   
CommonCommands = ["Fix It", "Fix Discrete", "What Can I Say",
                  "Select That", "Spell That", "Cut That", "Delete That", "Scratch That",
                  "Cap That", "\\All-Caps That", "\\No-Caps That", "Compound That",
                  "Hyphenate That", "Format That", "Copy That",
                  "Resume With",
                  "Next", "Forward", "Previous", "Back", "Left", "Right", "Top", "Bottom",
                  "Beginning", "End",
                  "Word", "Words", "Character", "Characters", "Line", "Lines", "Paragraph","Paragraphs",
                  "Select Line", "Go to Top", "Go to Bottom",
                  "Insert Before That", "Insert After That", "Select Again",
                  "Choose 1", "Choose 2", "Choose 3",
                  "Choose 4", "Choose 5", "Choose 6",
                  "Choose 7", "Choose 8", "Choose 9", "Choose 10",
                  "Select 1", "Select 2", "Select 3",
                  "Select 4", "Select 5", "Select 6",
                  "Select 7", "Select 8", "Select 9", "Select 10",
                  "Switch to", "Start"]

SpecialTraining={
    'Numbers':          Numbers,
    'AppStarts':        AppStarts,
    'TeXFundamentals':  TeXFundamentals,
    'CommonCommands':   CommonCommands,
    'Alphabet':         D_.Alphabet,
    'ICAlphabet':       D_.ICAlphabet,        
    }



import re
import sys
import copy
from dtactions.unimacro import utilsqh

reNote = re.compile(r"""([a-g](?:as|es|is|s|f)?)  # the notename also with flat (bf == b flat, english)
                        ([,']*)                 # elevation (octave up/down                        
                        ([0-9]*[.]*)            # duration, including augmentation dots
                        (.*)$                   # rest (assume no spaces)
                    """, re.VERBOSE+re.UNICODE)

reRest = re.compile(r"""([rRsS])  # the name of the rest
                        ##([,']*)                 # elevation (octave up/down                        
                        ([0-9]*[.]*)            # duration, including augmentation dots
                       ### (.*)$                   # rest (assume no spaces)
                    """, re.VERBOSE+re.UNICODE)
reBackslashedWord = re.compile(r"""([\\]\w+)""")

# for analyseString:
reSplitInWords = re.compile(r'(\s+)')  # split on whitespace
reIsWhiteSpace = re.compile(r'\s')



class LyNote:
    def __init__(self, s):
        self.originalInput = s
        self.setVariables(s)
    
    def __len__(self):
        return len(str(self))
    
        
    def setVariables(self, s, recursive=None):
        """parse s and set self.note etc."""
        
        m = reNote.match(s)
        n = reRest.match(s)
        self.note = self.elevation = self.duration = ""
        rest = s
        self.additions = ""
        self.backslashed = None
        if m:
            self.note = m.group(1)
            self.elevation = m.group(2)
            self.duration = m.group(3)
            rest = m.group(4)
            if rest:
                self.additions, self.backslashed = self.orderRest(rest)
                # backslashed is a list of \melisma etc.
        elif n:
            self.note = n.group(1)
            self.duration = n.group(2)
            self.elevation = self.rest = ""
            
        else:
            # can be incomplete, add fake note in front and do again
            if recursive:
                print('recursive call of setVariables failed, s: %s'% s)
                return
            # print 'try recursive call of incomplete note: %s'% s
            sFake = 'c' + s
            self.setVariables(sFake, recursive=1)
            self.note = ""


    def isNote(self):
        return ( self.note != "" )
    
    def __str__(self):
        if self.backslashed:
            backslashed = " " + ' '.join(self.backslashed)
        else:
            backslashed = ""
        return "%s%s%s%s%s"% (self.note, self.elevation, self.duration, self.additions, backslashed)
    
    def __repr__(self):
        if self.backslashed:
            backslashed = " " + ' '.join(self.backslashed)
        else:
            backslashed = ""
        return "lynote: %s%s%s%s%s"% (self.note, self.elevation, self.duration, self.additions, backslashed)
    
    def getElevation(self):
        """return as int the elevation
        , = -1, ,, = -2, ' = 1, '' = 2 and "" = 0
        """
        if self.elevation == "":
            return 0
        if self.elevation.find(",") >= 0:
            return -len(self.elevation)
        if self.elevation.find("'") >= 0:
            return len(self.elevation)
        raise ValueError('getElevation: no valid elevation: "%s" (note: "%s"'% (self.elevation, self))
    
    def setElevation(self, elevation):
        """set the elevation string from the numeric elevation value
        0 = "", -1 = ",", 2 = "''" etc
        """
        if elevation == 0:
            return ""
        if elevation > 0:
            return "'"*elevation
        if elevation < 0:
            return ","*-elevation
        
        
    
    def orderRest(self, rest):
        if rest is None:
            return "", None
        
        if rest.find('\\') >= 0:
            additions = ""
            backslashed = []
            restList = reBackslashedWord.split(rest)
            for item in restList:
                if not item:
                    continue
                if reBackslashedWord.match(item):
                    backslashed.append(item)
                else:
                    additions += item

            return additions, backslashed
        return rest, None
        
    
    
    def updateNote(self, note):
        if type(note) in (bytes, str):
            note = LyNote(note)
        
        for attr in ['note', 'duration']:
            value = getattr(note, attr)
            if value:
                setattr(self, attr, value)
        if note.elevation:
            nElv = note.getElevation()
            orgElv = self.getElevation()
            newElv = orgElv + nElv
            self.elevation = self.setElevation(newElv)
            
        if note.additions:
            if self.additions:
                self.additions += note.additions
            else:
                self.additions = note.additions
        if note.backslashed:
            if self.backslashed:
                self.backslashed.extend(note.backslashed)
            else:
                self.backslashed = copy.copy(note.backslashed)

def analyseString(S):
    """split string in "words", leave the whitespace and return a second list of the note indexes
    
    return wordsList, notesIndexes, both being a list. the length of notesIndexes gives the number
    of notes found.
    """
    wordsList = [_f for _f in reSplitInWords.split(S) if _f]
    outputWords = []
    notesIndexes = []
    lyn = None
    inComment = 0
    peekwords = utilsqh.peekable(wordsList)
    index = 0

    for w in peekwords:
        if w == "%":
            comment = [w]
            while peekwords.peek() != '\n':
                comment.append(next(peekwords))
            outputWords.append(''.join(comment))
            index += 1
        
        if w and w[0] in 'abcdefgrR':
            lyn2 = LyNote(w)
            if lyn2.isNote():
                # print 'valid note: %s'% lyn2
                
                notesIndexes.append(index)
                outputWords.append(lyn2)
                index += 1
            else:
                print('not a note: %s'% w)
        elif w and w[0].strip():
            outputWords.append(w)
            index += 1
        else:
            # whitespace/text
            intermediate = []
        
            while 1:
                intermediate.append(w)
                peekword = peekwords.peek()
                if peekword is peekwords.sentinel:
                    outputWords.append(''.join(intermediate))
                    break
                if LyNote(peekword).isNote():
                    # flush:
                    outputWords.append(''.join(intermediate))
                    index += 1
                    break
                # else:
                #     print 'strange: peekword: %s'% repr(peekword)
                w = next(peekwords)

                
    return outputWords, notesIndexes
                    
def getNotesFromWordsList(n, wordsList, notesIndexes, reverse=False):
    """get exactly n notes from list, and return the start and end
    
    n = number of notes to be found
    wordsList: list of words, including notes
    notesIndexes: indexes of words that are notes
    reverse: with previous Notes, reverse the lists and back
    return:
    wordsList: the list of n notes and intermediate words
    notesIndexes: the list of indexes of the words, length exactly n
    beforeNotes: string that precedes the first note // is after the last note (reverse)
    afterNotes: string that follows the last note    // precedes the first note (reverse)
    afterNoteComplete: True if afterNotes extends until the next note  //
                                     // or until the previous note (reverse)
                                     
    if reverse == True, the last three variables should be interpreted in the opposite direction.
    
    """
    wordsList = copy.copy(wordsList)
    notesIndexes = copy.copy(notesIndexes)
    if reverse:
        lenw = len(wordsList)
        notesIndexes.reverse()
        wordsList.reverse()
        notesIndexes = [lenw-i-1 for i in notesIndexes]
    
    notesFound = len(notesIndexes)
    takeFromList = notesIndexes[n-1] + 1
    if notesFound == n:
        # string after last note:
        afterNotesList = wordsList[takeFromList:]
        afterNoteComplete = False  # not sure if reach until next note
    else:
        # take after n'th note until next note:
        takeAfterNotes = notesIndexes[n]
        afterNotesList = wordsList[takeFromList:takeAfterNotes]
        afterNoteComplete = True
    # print'afterNotes: %s,complete: %s'% (repr(afterNotes), afterNoteComplete) 
    wordsList = wordsList[:takeFromList]
    notesIndexes = notesIndexes[:n]
    firstNoteAt = notesIndexes[0]

    if firstNoteAt == 0:
        beforeNotesList = []
    else:
        # the string before the notes list:
        beforeNotesList = wordsList[:firstNoteAt]
        # print 'string before first note: %s'% beforeNotes
        wordsList = wordsList[firstNoteAt:]
        notesIndexes = [i-firstNoteAt for i in notesIndexes]
    # print 'notesIndexes: %s'% notesIndexes

    # reverse back:
    if reverse:
        lenw = len(wordsList)
        notesIndexes.reverse()
        wordsList.reverse()
        notesIndexes = [lenw-i-1 for i in notesIndexes]
        beforeNotesList.reverse()
        afterNotesList.reverse()
    
    afterNotes = ''.join(afterNotesList)
    beforeNotes = ''.join(beforeNotesList)

    return wordsList, notesIndexes, beforeNotes, afterNotes, afterNoteComplete
          
def join(Input, separator=''):
    """join all things (recursively if necessary)

    """
    if len(Input) == 0:
        return ''
    if type(Input) in (bytes, str):
        return Input
    else:
        return separator.join(str(i) for i in Input)



if __name__ == '__main__':
    
    #for s in ["", "(", r"(\break\melisma", r"\break\(-^2"]:
    #    mrest = reBackslashedWord.split(s)
    #    print 's: %s, mrest: %s'% (s, mrest)
    
    #for s in ["g,8.", "a", "cis'("]:
    #    lyn = LyNote(s)   
    #    print 'note: "%s", elevation: "%s", duration: "%s", additions: "%s"'% (lyn.note, lyn.elevation, lyn.duration, lyn.additions)
    #    print 'input: "%s", str: "%s", repr: "%s"'% (s, lyn, repr(lyn))
    #    lyn.updateNote("a")
    #    print 'note updated to a: %s'% lyn
    #    lyn.updateNote(r"c8.\(")
    #    print 'note updated to c 8. and \\(: %s'% lyn
    print('melisma: =============================================')
    for s in ["r2", r"g,8.\melisma", r"a\melisma"]:
        lyn = LyNote(s)
        print('note: "%s", elevation: "%s", duration: "%s", additions: "%s"'% (lyn.note, lyn.elevation, lyn.duration, lyn.additions))
        print('input: "%s", str: "%s", repr: "%s"'% (s, lyn, repr(lyn)))
        lyn.updateNote("a")
        print('note updated to a: %s'% lyn)
        lyn.updateNote(r"c8.\(")
        print('note updated to c 8. and \\(: %s'% lyn)
        lyn = LyNote(s)
        
        # join tests:
        print(join('abc'))
        print(join(['a', 'a', 'c']))
        print(join(['a', 'b'], separator='xxx'))


#
# Python Macro Language for Dragon NaturallySpeaking
#   (c) Copyright 1999 by Joel Gould
#   Portions (c) Copyright 1999 by Dragon Systems, Inc.
#
#   This script performs some tests of natlinkutilsqh functions.
#   The DragonPad window is used, and killed afterwards
#
#   NaturallySpeaking should be running with nothing in the DragonPad window
#   (that you want to preserve) before these tests are run.
#
from dtactions.unimacro import unimacroutils
from natlinkcore import natlinkutils
from dtactions.unimacro import unimacroactions as actions
action = actions.doAction

import unittest
import types
import TestCaseWithHelpers


##class UnimacroUtilsTest(TestCaseWithHelpers.TestCaseWithHelpers):
class UtilsTest(TestCaseWithHelpers.TestCaseWithHelpers):

    # run from dragonpad      
    def setUp(self):
        actions.doAction("BRINGUP dragonpad")

    def tearDown(self):
        actions.doAction("BRINGUP dragonpad; KW")
        
    def log(self, mess):
        if type(mess) == list:
            mess = '\n'.join(mess)
        actions.doAction(str(mess))

    def test_Something_in_unimacro(self):
        self.log('testing something')
        actions.doAction("W")
        lang = unimacroutils.getLanguage()
        self.assert_equal("enx", lang, "testing should be done from an English speech profile, not: %s"% lang)

    def test_matchModule(self):
        """check if dragonpad is reported as top window

        and the open dialog as child window

         behaviour changed in python version 2.3.4!!
        """        
        progInfo = unimacroutils.getProgInfo()
        self.log("progInfo: %s"% repr(progInfo))
        actions.doAction("LW")
        ## changed in 2013 from 3 to 4 items:
        self.assert_equal(4, len(progInfo), "progInfo should be tuple of 4")
        self.assert_equal(u'top', progInfo[2], "DragonPad should be a top window")
        actions.doAction("<<fileopen>>")
        progInfo = unimacroutils.getProgInfo()
        actions.doAction("{esc}")
        self.assert_equal(4, len(progInfo), "progInfo should be tuple of 4")
        self.assert_equal(u'child', progInfo[2], "DragonPad open dialog should be a child window")
        
    def test_actionsIniFilePresent(self):
        """see if the ini file of actions is there
        can be used to manually test the copying of sample ini files in different configurations
        """
        pass
        ## think about testing the import of actions module in different config settings
        # eg with and without Unimacro enabled, also with different possibilities of sample ini files

        
# no main statement, run from command in _unimacrotest.py


##############################################################################
# VoiceCode, a programming-by-voice environment
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# (C)2000, National Research Council of Canada
#
##############################################################################

"""Various utility functions"""

import getopt
import os
import re
import stat
import sys
import time
import types
import sys
if sys.platform =='win32':
    global winsound
    import winsound

import os.path

import vc_globals

def dict_merge(dict1, dict2):
    """Merges two dictionaries
    
    Merges *{ANY: ANY} dict1* and *{ANY: ANY} dict2* and returns the
    result *{ANY: ANY} dict3*.

    If a key exists in both *dict1* and *dict2*, the value in *dict1*
    will be used for *dict3*
    """
    dict3 = dict2
    for an_item in dict1.items():
        key, val = an_item
        dict3[key] = val
    return dict3



def gopt(opt_defs, cmd=None):
    """High level wrapper around *getopt.getop*.

    *removes first argument from *cmd* when parsing from *sys.argv*
        
    *returned options are stored in a dictionary

    *dashes ('-') are removed from the option name in that dictionary

    *returns None and outputs error messages if invalid option

    *allows to define default values for options
        
    **INPUTS**
        
    *[STR]* cmd=sys.argv[1:] -- list of options and arguments to be parsed. 
        
    *[STR, ANY, ...]* opt_defs -- defines the valid options (short and
    long). The list is an alternate sequence of option name and
    default value. If the name ends with *=*, it means the option
    requires a value. If the name is a single letter, it's a short
    option. The defaul value is compulsory, even for options that
    don't require a value (can be used to set the switch to on or
    off by default).
        

    **OUTPUTS**
        
    *opts, args* -- *opts* is a dictionary of options names and
    values. *args* is the list of arguments.
    """
    if cmd == None:
        cmd = sys.argv[1:]
    opt_dict = {}
    args = []    
    
    #
    # Set default values of options
    #
    index = 0
    while (index < len(opt_defs)):
        opt_name = opt_defs[index]
        opt_default = opt_defs[index + 1]
        opt_name = re.match('^(.*?)(=*)$', opt_name).groups()[0]        
        opt_dict[opt_name] = opt_default
        index = index + 2
#    print '-- util.gopt: initialised opt_dict=%s' % repr(opt_dict)


    #
    # Set options specifications to be used by getopt.
    #
    short_opts = ''
    long_opts = []

    requires_val = {}
    is_long = {}    
    index = 0
    while (index < len(opt_defs)):
        opt_name = opt_defs[index]
        opt_default = opt_defs[index + 1]
        index = index + 2
        match = re.match('^(.)(.*?)(=*)$', opt_name)
        opt_name = match.group(1)+ match.group(2)
        if (match.group(2) != ''):
            is_long[opt_name] = 1
        if (match.group(3) != ''):
            requires_val[opt_name] = 1
            
        if is_long.has_key(opt_name):
            long_opts = long_opts + [opt_name + match.group(3)]
        else:
            short_opts = short_opts + opt_name
            if requires_val.has_key(opt_name):
                short_opts = short_opts + ':'

    #
    # Parse the command line options
    #
#    print '-- util.gopt: calling getopt with cmd=%s, short_opts=%s, long_opts=%s' % (repr(cmd), repr(short_opts), repr(long_opts))                
    options, args = getopt.getopt(cmd, short_opts, long_opts)    
#    print '-- util.gopt: options=%s, args=%s' % (repr(options), repr(args))


    #
    # Assign parsed values to the options dictionary.
    #
#    print '-- util.gopt: is_long=%s, requires_val=%s' % (repr(is_long), repr(requires_val))
    for an_opt in options:
        opt_name = an_opt[0]
        a_match = re.match('^(-*)(.*)$', opt_name)
        opt_name = a_match.group(2)
        if not requires_val.has_key(opt_name):
            #
            # getopt.getopt returns None as the value for BOOLEAN options
            # but we want it to be 1, otherwise it makes it look like the
            # options was off
            #
            # In getopt.getopt, that didn't matter because the mere presence
            # of the option name indicates it is on.
            #
            # In util.gopt, all options have an entry in the returned
            # dictionary, and its value indicates whether it's on or off
            #
            opt_val = 1
        else:
            opt_val = an_opt[1]
        opt_dict[opt_name] = opt_val

#    print "-- gopt: opt_dict=%s, args=%s" % (str(opt_dict)    , str(args))        
    return opt_dict, args

###############################################################################
# list processing
###############################################################################

def remove_occurences_from_list(item, list, max_occurs=None):
   num_found = 0
   new_list = []
   ii = 0
   for ii in range(len(list)):
      if list[ii] == item:
         num_found = num_found + 1
         if max_occurs != None and num_found >= max_occurs:
            break
      else:
         new_list.append(list[ii])
      
   ii_rest = ii + 1   
   if ii_rest < len(list):
      new_list = new_list + list[ii_rest:]

      
   return new_list


###############################################################################
# file system
###############################################################################


def last_mod(f_name):
    """Returns the time at which a file was last modified.

    *STR f_name* is the path of the file.

    if *f_name* doesn't exist, returns 0.
    """

    try:
        stats = os.stat(f_name)
        time = stats[stat.ST_MTIME]
    except OSError:
        time = 0
        
    return time


###############################################################################
# For redirecting STDIN
###############################################################################

#
# Use this temporary file to send commands to the process' own stdin
#
redirected_stdin_fname = vc_globals.tmp + os.sep + 'user_input.dat'
redirected_stdin = open(redirected_stdin_fname, 'w')

def stdin_read_from_string(string):
    """Redirects process' own stdin so it reads from a string

    **INPUTS**
    
    *[STR] string* -- String from which to read stdin

    **OUTPUTS**

    *FILE old_stdin* -- Stream that stdin was originally assigned to.
    """ 

    global redirected_stdin, redirected_stdin_fname
    
    old_stdin = sys.stdin

    #
    # Close temporary file in case it was opened by a previous call to
    # stdin_read_from_string
    #
    redirected_stdin.close()

    #
    # Write the string to temporary file
    #
    redirected_stdin = open(redirected_stdin_fname, 'w')
    redirected_stdin.write(string)
    redirected_stdin.close()

    #
    # Open temporary file and assign it to stdin
    #
    redirected_stdin = open(redirected_stdin_fname, 'r')
    sys.stdin = redirected_stdin

    return old_stdin

###########################################################################
# Identifying instances of basic types
###########################################################################

def islist(instance):
    """Returns true iif *instance* is a list."""

    return isinstance(instance, types.ListType)
    
def istuple(instance):
    """Returns true iif *instance* is a tuple."""

    return isinstance(instance, types.TupleType)

def issequence(instance):
    return islist(instance) or istuple(instance)

def isfunction(instance):
    return isinstance(instance, types.FunctionType)
   

###############################################################################
# path manipulation 
###############################################################################

def full_split(path):
    head, tail = os.path.split(path)
    if head == '': return [tail]
    if head == path: return [path]
    l = full_split(head)
    l.append(tail)
    return l

def find_common(a, b):
    n = min(len(a), len(b))
    for i in range(n):
        if a[i] != b[i]:
            return i
    return n

def remove_common(first, second):
    f = full_split(first)
    s = full_split(second)
    n = find_common(f,s)
    return f[n:]

def common_Path(first, second):
    f = full_split(first)
    s = full_split(second)
    n = find_common(f,s)
    return f[:n]

def relative_name(path, head, prefix=''):
    common = common_Path(path, head)
    whole_head = full_split(head)
    if (whole_head != common):
        return path
    p = []
    if (prefix != ''):
        p.append(prefix)
    p.extend(remove_common(path,head))
    return apply(os.path.join, p)

def replace_all_within_VCode(text, prefix='%VCODE_HOME%' ):
    return text.replace(vc_globals.home, prefix)

def within_VCode(path):
#    return relative_name(path, vc_globals.home)
    return relative_name(path, vc_globals.home, prefix='%VCODE_HOME%')
#    return relative_name(path, vc_globals.home, 
#        prefix = os.path.basename(vc_globals.home))


###############################################################################
# miscellaneous
###############################################################################


def bell(length=3):
    """Plays a bell sound for a time proportional to INT length.
    """
    
    bell_string = ''
    for ii in range(length):
       bell_string = bell_string + '\a'
# use sys.__stderr__ so that the user will hear the bell even if stderr
# is redirected by regression testing
    sys.__stderr__.write(bell_string)
    sys.__stderr__.flush()

# AD: Uncomment once we have upgraded to wxPython 2.5.    
#import wx
#def wxPython_is_unicode_build():
#   if 'unicode' in wx.PlatformInfo:
#      print "wxPython was built in Unicode mode"
#      return True
#   else:
#      print "wxPython was built in ANSI mode"
#      return False




#
# Python Macro Language for Dragon NaturallySpeaking
#   (c) Copyright 1999 by Joel Gould
#   Portions (c) Copyright 1999 by Dragon Systems, Inc.
#
# unittestDocstringGrammar.py
#   This script performs some basic tests of the NatLink but
#   with the DocstringGrammar subclass.
#   This also tests features of IniGrammar, BrowsableGrammar and GrammarX
#   being the chain of grammars leading up to GrammarBase
#
# run from a (preferably clean) US user profile, easiest from IDLE.
# do not run from pythonwin. 
#
import sys
import unittest
import types
import os
import os.path
import time
import traceback        # for printing exceptions
import TestCaseWithHelpers
import natlink
from natlinkcore import loader
from natlinkcore import natlinkstatus
from natlinkcore.gramparser import GrammarError, GrammarSyntaxError

status = natlinkstatus.NatlinkStatus()

from natlinkcore import natlinkutils
from dtactions.unimacro import unimacroutils
from dtactions.unimacro import unimacroutils
import unimacro.natlinkutilsbj as natbj
from dtactions.unimacro.unimacroactions import doAction as action
from dtactions.unimacro.unimacroactions import doAction as action

class TestError(Exception):
    pass
ExitQuietly = 'ExitQuietly'

def getBaseFolder(globalsDict=None):
    """get the folder of the calling module.

    either sys.argv[0] (when run direct) or
    __file__, which can be empty. In that case take the working directory
    """
    globalsDictHere = globalsDict or globals()
    baseFolder = ""
    if globalsDictHere['__name__']  == "__main__":
        baseFolder = os.path.split(sys.argv[0])[0]
        print('baseFolder from argv: %s'% baseFolder)
    elif globalsDictHere['__file__']:
        baseFolder = os.path.split(globalsDictHere['__file__'])[0]
        print('baseFolder from __file__: %s'% baseFolder)
    if not baseFolder or baseFolder == '.':
        baseFolder = os.getcwd()
        print('baseFolder was empty, take wd: %s'% baseFolder)
    return baseFolder

thisDir = getBaseFolder(globals())

natconnectOption = 0 # or 1 for threading, 0 for not. Seems to make difference
                     # with spurious error (if set to 1), missing gotBegin and all that...
logFileName = os.path.join(thisDir, "testresult.txt")

#---------------------------------------------------------------------------
# These tests should be run after we call natConnect
# no reopen user at each test anymore..
# no default open window (open window will be the calling program)
# default .ini files pop up when you first run this test. just ignore them.
class UnittestDocstringGrammar(TestCaseWithHelpers.TestCaseWithHelpers):
    def setUp(self):
        if not natlink.isNatSpeakRunning():
            raise TestError('NatSpeak is not currently running')
        self.connect()
        self.user = natlink.getCurrentUser()[0]
        self.setMicState = "off"

    def tearDown(self):
        try:
            # give message:
            self.setMicState = "off"
            # kill things
            self.clearTestFiles()
        finally:
            self.disconnect()

        
    def connect(self):
        # start with 1 for thread safety when run from pythonwin:
        natlink.natConnect(natconnectOption)

    def disconnect(self):
        natlink.natDisconnect()
        
    def log(self, t):
        # only log to file:
        log(t)

    def clearTestFiles(self):
        """remove .py and .pyc files from the natlinkmain test

        """
        return
        # try to remove __main__.ini, but gives unexpected windows popping up
        #userdir = status.getUnimacroUserDirectory()
        #language = status.getLanguage()
        #print 'userdir: %s, language: %s'% (userdir, language)
        #userinidir = os.path.join(userdir, '%s_inifiles'%language)
        #if os.path.isdir(userinidir):
        #    mainfile = os.path.join(userinidir, '__main__.ini')
        #    if os.path.isfile(mainfile):
        #        os.remove(mainfile)
        #else:
        #    raise OSError("clearTestFiles, should be a valid directory: %s"% userinidir)

    def wait(self, t=1):
        time.sleep(t)


    #---------------------------------------------------------------------------
    # This utility subroutine executes a Python command and makes sure that
    # an exception (of the expected type) is raised.  Otherwise a TestError
    # exception is raised

    def doTestForException(self, exceptionType,command,localVars={}):
        try:
            exec(command,globals(),localVars)
        except exceptionType:
            return
        raise TestError('Expecting an exception to be raised calling '+command)

    def doTestActiveRules(self, gram, expected):
        """gram must be a grammar instance, sort the rules to be expected and got
        """
        expected.sort()
        got = gram.activeRules
        got.sort()
        self.assertEqual(expected, got,
                         'Active rules not as expected:\nexpected: %s, got: %s'%
                         (expected, got))

    def doTestFuncReturn(self, expected,command,localVars=None):
        # account for different values in case of [None, 0] (wordFuncs)
        if localVars == None:
            actual = eval(command)
        else:
            actual = eval(command, globals(), localVars)

        if actual != expected:
            time.sleep(1)
        self.assertEqual(expected, actual, 'Function call "%s" returned unexpected result\nExpected: %s, got: %s'%
                          (command, expected, actual))
    
    def testGrammarActivateRules(self):
        """test a simple grammar with three exported rules.
        
        test the activation/deactivation of these rules, like in unittestNatlink
        (the testGrammar function)
        """
        class TestGrammar(natbj.DocstringGrammar):
            gramSpec = '<four> exported = rule four;'
            
            def __init__(self):
                natbj.DocstringGrammar.__init__(self)
                self.resetExperiment()

            def resetExperiment(self):
                self.sawBegin = 0
                self.recogType = None
                self.words = []
                self.fullResults = []
                self.error = None

            def gotBegin(self,moduleInfo):
                if self.sawBegin > nTries:
                    self.error = 'Command grammar called gotBegin twice'
                self.sawBegin += 1
                if moduleInfo != natlink.getCurrentModule():
                    self.error = 'Invalid value for moduleInfo in GrammarBase.gotBegin'

            def gotResultsObject(self,recogType,resObj):
                if self.recogType:
                    self.error = 'Command grammar called gotResultsObject twice'
                self.recogType = recogType
            def rule_one(self, words):
                'rule one <two>'
                pass
            def subrule_two(self, words):
                'subrule two'
                pass
            def rule_three(self, words):
                'rule three'
                pass

            def gotResults(self,words,fullResults):
                if self.words:
                    self.error = 'Command grammar called gotResults twice'
                self.words = words
                self.fullResults = fullResults

            def checkExperiment(self,sawBegin,recogType,words,fullResults):
                if self.error:
                    raise TestError(self.error)
                if self.sawBegin != sawBegin:
                    raise TestError('Unexpected result for GrammarBase.sawBegin\n  Expected %d\n  Saw %d'%(sawBegin,self.sawBegin))
                if self.recogType != recogType:
                    raise TestError('Unexpected result for GrammarBase.recogType\n  Expected %s\n  Saw %s'%(recogType,self.recogType))
                if self.words != words:
                    raise TestError('Unexpected result for GrammarBase.words\n  Expected %s\n  Saw %s'%(repr(words),repr(self.words)))
                if self.fullResults != fullResults:
                    raise TestError('Unexpected result for GrammarBase.fullResults\n  Expected %s\n  Saw %s'%(repr(fullResults),repr(self.fullResults)))
                self.resetExperiment()
       
        testActiveRules = self.doTestActiveRules
        testForException = self.doTestForException
        testGram = TestGrammar()
        
        expGramSpec =  \
'''<one> exported = rule one <two>;
    <two> = subrule two;
<three> exported = rule three;
<four> exported = rule four;'''
        self.assert_equal(expGramSpec, testGram.gramSpec,"gramspec not as expected")
        
        testGram.load(testGram.gramSpec)
        testGram.activateAll()
        testActiveRules(testGram, ['one', 'three', 'four'])

        testGram.deactivateAll()
        testActiveRules(testGram, [])
        prev = None
        for rule in 'three', 'four', 'one', 'four', 'three':
            # activate:
            testGram.activate(rule)
            if prev:
                expList = [prev, rule]
            else:
                expList = [rule]
            testActiveRules(testGram, expList)
            # activate after all active:
            testGram.activateAll()
            testForException(GrammarError, "testGram.activate('%s')"% rule, locals())
            # activate after all unactive:
            testGram.deactivateAll()
            testGram.activate(rule)
            testActiveRules(testGram, [rule])
            prev = rule

        rule = 'one'
        testGram.activate(rule)
        testForException(GrammarError, "testGram.activate('%s')"% rule, locals())
        testGram.deactivateAll()
        testActiveRules(testGram, [])

        for SET in (['one', 'three', 'four'], ['three'], ['one', 'three', 'four'], ['one', 'three']):
            testGram.activateSet(SET)
            testActiveRules(testGram, SET)
            ##with original version of natlinkutils.py you get:
            ##AssertionError: Active rules not as expected:
            ##expected: ['three'], got: ['one', 'three']
            ##fix around line 420 (copy.copy) in natlinkutils.py, QH
    
        ## test exceptlist feature:
        ## ['one', 'three', 'four'] are the exported rules
        testGram.activateAll(exceptlist=['one'])
        testActiveRules(testGram, ['three', 'four'])
        testGram.activateAll(exceptlist=None)
        testActiveRules(testGram, ['one', 'three', 'four'])
        testGram.deactivateAll()
        testActiveRules(testGram, [])
        # extra in exceptlist does not matter:
        testGram.activateAll(exceptlist=['one', 'two', 'three', 'four', 'five'])
        testActiveRules(testGram, [])
        testGram.activateAll(exceptlist=['three', 'four'])
        testActiveRules(testGram, ['one'])
        testGram.activateAll(exceptlist=['one', 'three'])
        testActiveRules(testGram, ['four'])
        
        

        # try a few illegal grammars to make sure they are reported properly (we
        # already tested the grammar parser so this does not have to be
        # exhaustive)
        testGram.unload()
        testForException(SyntaxError,"testGram.load('badrule;')",locals())
        testForException(GrammarError,"testGram.load('<rule> = hello;')",locals())

        # most calls are not legal before load is called (successfully)
        testForException(natlink.NatError,"testGram.gramObj.activate('start',0)",locals())
        testForException(natlink.NatError,"testGram.gramObj.deactivate('start')",locals())
        testForException(natlink.NatError,"testGram.gramObj.setExclusive(1)",locals())
        testForException(natlink.NatError,"testGram.gramObj.emptyList('list')",locals())
        testForException(natlink.NatError,"testGram.gramObj.appendList('list','word')",locals())
        # clean up
        testGram.unload()
        
        

    #def testGrammarNumberFunctionsWithoutNumbersIni(self):
    #    """test the functions that convert numbers lists
    #    
    #    when the spokenforms.ini has been removed (temporarily)
    #    """
    #    class TestGrammar(natbj.DocstringGrammar):
    #        name = "testgrammarone"
    #        def __init__(self):
    #            natbj.DocstringGrammar.__init__(self)
    #        def rule_one(self, words):
    #            'rule one'
    #            pass
    #   
    #    testGram = TestGrammar()
    #    
    #    expGramSpec =  '''<one> exported = rule one;'''
    #    self.assert_equal(expGramSpec, testGram.gramSpec,"gramspec not as expected")
    #    
    #    # test the numbers convert function
    #    numIni = testGram.getNumbersInifile()
    #    language = 'enx'
    #    self.assert_equal(language, testGram.language, "Testing should take place with 'enx' speech profile, not: %s"% testGram.language)
    #    hundredGot = numIni.getList(language + "_prefixes", "hundred")
    #    expHundredList = ["hundred", "one hundred", "a hundred"]
    #    self.assert_equal(expHundredList, hundredGot, "list of 'hundred prefixes should be as expected")
    #    result = testGram.getSpokenFormsAboveHundred(102, numIni)
    #    expResult = ["hundred two", "one hundred two", "a hundred two"]
    #    self.assert_equal(expResult, result, "list for 102 not as expected")
    #    result = testGram.getSpokenFormsAboveHundred(335, numIni)
    #    expResult = ['three hundred thirty-five', 'three thirty-five']
    #    self.assert_equal(expResult, result, "list for 335 not as expected")

    def testGrammarNumberFunctions(self):
        """test the functions that convert numbers lists
        
        """
        class TestGrammar(natbj.DocstringGrammar):
            name = "testgrammarnumbers"
            def __init__(self):
                natbj.DocstringGrammar.__init__(self)
            def initialize(self):
                self.numresult = None
                self.switchOn()
            def rule_one(self, words):
                'rule one <n1-10>'
                pass
            def rule_two(self, words):
                'rule two  <numbers2to18>'
                pass
            def rule_three(self, words):
                'rule three  <n2to20>'
                pass
            def rule_four(self, words):
                'rule four  <numbers1-100>'
                self.numresult = self.getNumberFromSpoken(words[0])
            def rule_five(self, words):
                'rule five <n0-5>'
                self.numresult = self.getNumberFromSpoken(words[0])
            def checkExperiment(self, expected, testInfo):
                self.test.assert_equal(expected, self.numresult, testInfo)
            def gotBegin(self, modInfo):
                self.numresult = None # reset
                
                
        testGram = TestGrammar()
        testGram.test = self

        def doTestNumbersRecognition(words, expected, info):
            natlink.recognitionMimic(words)
            testGram.checkExperiment(expected, info)
        expGramSpec =  \
'''<one> exported = rule one <n1-10>;
<two> exported = rule two  <numbers2to18>;
<three> exported = rule three  <n2to20>;
<four> exported = rule four  <numbers1-100>;
<five> exported = rule five <n0-5>;'''
        self.assert_equal(expGramSpec, testGram.gramSpec,"gramspec not as expected")
        language = 'enx'
        self.assert_equal(language, testGram.language, "Testing should take place with 'enx' speech profile, not: %s"% testGram.language)
        doTestNumbersRecognition(["rule", "one", "two"], 2, "testing grammar rule one with number 2")
        
        


def log(t):
    """log to print and file if present

    note print depends on the state of natlink: where is goes or disappears...
    I have no complete insight is this, but checking the logfile afterwards
    always works (QH)
    """
    print(t)
    if logFile:
        logFile.write(t + '\n')
    
#---------------------------------------------------------------------------
# run
#
# This is the main entry point.  It will connect to NatSpeak and perform
# a series of tests.  In the case of an error, it will cleanly disconnect
# from NatSpeak and print the exception information,
def dumpResult(testResult, logFile):
    """dump into 
    """
    if testResult.wasSuccessful():
        mes = "all succesful"
        logFile.write(mes)
        return
    logFile.write('\n--------------- errors -----------------\n')
    for case, tb in testResult.errors:
        logFile.write('\n---------- %s --------\n'% case)
        logFile.write(tb)
        
    logFile.write('\n--------------- failures -----------------\n')
    for case, tb in testResult.failures:
        logFile.write('\n---------- %s --------\n'% case)
        logFile.write(tb)

    


logFile = None

def run():
    global logFile, natconnectOption
    logFile = open(logFileName, "w")
    log("log messages to file: %s"% logFileName)
    log('starting unittestNatlink')
    # trick: if you only want one or two tests to perform, change
    # the test names to her example def test....
    # and change the word 'test' into 'tttest'...
    # do not forget to change back and do all the tests when you are done.
    suite = unittest.makeSuite(UnittestDocstringGrammar, 'test')
##    natconnectOption = 0 # no threading has most chances to pass...
    log('\nstarting tests with threading: %s\n'% natconnectOption)
    result = unittest.TextTestRunner().run(suite)
    dumpResult(result, logFile=logFile)
    
    logFile.close()

if __name__ == "__main__":
    run()
#


"""template for _folders grammar, remember function.
This template is filled with the actual values and fired. It asks for a spoken form to "remember a given file", this value
is put in the _folders.ini config file.
"""
#pylint:disable=W0621
import time
import PySimpleGUI as sg      
from dtactions.unimacro import inivars

prompt = """Remember in Unimacro _folders grammar"""  # readable text
text = """Remember website for future calling?

- https://pypi.org -

Please give a spoken form for this website and choose OK; or Cancel..."""          # input text, the key of the 
inifile = "C:/Users/Gebruiker/Documents/unimacro_qh/enx_inifiles/_folders.ini"
section = "websites"
value = "https://pypi.org"
title = "test"
default = "pypi"
pausetime = 3  # should be replaced by 0 or a positive int value



def InputBox(text, prompt, title, default):
    """the dialog, which returns the wanted spoken form"""
    layout = [[sg.Text(prompt)],      
                     [sg.InputText(default)],      
                     [sg.OK(), sg.Cancel()]]      
    
    window = sg.Window(title, layout)    
    
    _event, values = window.read()    
    window.close()
    
    return values[0]    

if __name__ == "__main__":
    result = InputBox(text, prompt, title, default)
    if result:
        key = result
        ini = inivars.IniVars(inifile)
        ini.set(section, key, value)
        ini.write()
        print(f'Wrote "{key} = {value}" to inifile')
        print('Call "edit folders" to edit or delete')
    else:
        print("Action canceled, no change of ini file")
    # if pausetime is given, launch as .py, otherwise launch as .pyw:
    if pausetime:
        # print "sleep: %s"% pausetime
        time.sleep(pausetime)
    